package com.bingce.controlnetwork.fragment;

import android.os.Bundle;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.annotation.WorkerThread;
import androidx.fragment.app.DialogFragment;
import androidx.fragment.app.FragmentManager;

import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory;
import com.bingce.controlapphelper.datasource.database.point.PointRecord;
import com.bingce.controlapphelper.util.PointUtil;
import com.bingce.controlapphelper.util.Tools;
import com.bingce.controlnetwork.R;
import com.bingce.controlnetwork.databinding.FragmentPointLayoutBinding;
import com.bingce.utils.SoftKeyUtils;
import com.bingce.utils.ThreadPoolUtil;
import com.rengwuxian.materialedittext.MaterialEditText;

import org.jetbrains.annotations.NotNull;

import java.util.Locale;

import blankj.utilcode.util.ToastUtils;

/**
 * 新建或编辑点
 */
public class PointFragment extends DialogFragment {

    private FragmentPointLayoutBinding mBinding;

    private PointFragment() {
    }

    @Override
    public void onCreate(@Nullable Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setStyle(STYLE_NORMAL, R.style.Dialog_FullScreen);
    }

    @Override
    public void onDestroyView() {
        super.onDestroyView();
        mBinding = null;
    }

    @Nullable
    @Override
    public View onCreateView(@NonNull @NotNull LayoutInflater inflater, @Nullable ViewGroup container, @Nullable Bundle savedInstanceState) {
        mBinding = FragmentPointLayoutBinding.inflate(inflater, container, false);
        return mBinding.getRoot();
    }

    @Override
    public void onViewCreated(@NonNull View view, @Nullable Bundle savedInstanceState) {
        super.onViewCreated(view, savedInstanceState);
        initListener();
        initData();
    }

    private void initListener() {

        mBinding.getRoot().setOnClickListener(v -> dismissPointFragment());

        mBinding.cancel.setOnClickListener(v -> dismissPointFragment());

        mBinding.next.setOnClickListener(v -> checkSave(isCanSave -> {
            if (!isCanSave) return;
            saveCurrnetPoint();
            startNextPointPre();
        }));

        mBinding.confirm.setOnClickListener(v -> checkSave(isCanSave -> {
            if (!isCanSave) return;
            saveCurrnetPoint();
            dismissPointFragment();
        }));

    }

    private void startNextPointPre() {
        runOnUI(() -> {
            //编号+1
            String indexString = string(mBinding.number);
            int index = 1;
            try {
                index = Integer.parseInt(indexString) + 1;
            } catch (Exception ignored) {
                ignored.printStackTrace();
            }
            mBinding.number.setText(String.format(Locale.getDefault(), "%03d", index));
            clearXYZ(mBinding.pointX, mBinding.pointY, mBinding.pointZ);
        });
    }

    @WorkerThread
    private void saveCurrnetPoint() {
        String pointName = pointName(mBinding.prefix, mBinding.number, mBinding.suffix);

        double x = doubleValue(mBinding.pointX);
        double y = doubleValue(mBinding.pointY);
        double z = doubleValue(mBinding.pointZ);

        if (isEdit()) {
            String pointId = getPointId();
            PointRecord pointRecord = SurveyorDatabaseFactory.instance.getPointDataSource().findByIdSync(pointId);
            pointRecord.name = pointName;
            pointRecord.x = x;
            pointRecord.y = y;
            pointRecord.z = z;
            pointRecord.code = "";
            pointRecord.remarks = "";
            SurveyorDatabaseFactory.instance.getPointDataSource().update(pointRecord);
        } else {
            PointRecord pointRecord = PointRecord.point(
                    getProjectId(), getPointType(), pointName, "", x, y, z, "", -1);
            SurveyorDatabaseFactory
                    .instance
                    .getPointDataSource()
                    .saveSync(pointRecord);
        }
    }

    /**
     * 初始化编辑数据
     */
    private void initData() {
        boolean isEdit = isEdit();

        mBinding.title.setText(isEdit ? R.string.edit_point : R.string.new_point);

        if (isEdit) {
            mBinding.number.setVisibility(View.GONE);
            mBinding.suffix.setVisibility(View.GONE);
            mBinding.next.setVisibility(View.GONE);

            setString(mBinding.prefix, getArguments(), KEY_POINT_NAME);
            String label = getString(R.string.point_name);
            mBinding.prefix.setFloatingLabelText(label);
            mBinding.prefix.setHint(label);

            setDouble(mBinding.pointX, getArguments(), KEY_POINT_X);
            setDouble(mBinding.pointY, getArguments(), KEY_POINT_Y);
            setDouble(mBinding.pointZ, getArguments(), KEY_POINT_Z);
        }
    }

    private void runOnUI(Runnable runnable) {
        if (getActivity() != null) {
            getActivity().runOnUiThread(runnable);
        }
    }

    private void clearXYZ(MaterialEditText x, MaterialEditText y, MaterialEditText z) {
        if (x != null) {
            x.setText(null);
        }
        if (y != null) {
            y.setText(null);
        }
        if (z != null) {
            z.setText(null);
        }
    }

    private String string(MaterialEditText editText) {
        if (editText == null || editText.getText() == null) {
            return "";
        }
        return editText.getText().toString();
    }

    private double doubleValue(MaterialEditText editText) {
        String string = string(editText);
        if (string.isEmpty()) {
            return 0;
        }
        return Double.parseDouble(string);
    }

    private void setString(MaterialEditText editText, Bundle args, String key) {
        if (editText == null || args == null) {
            return;
        }
        editText.setText(args.getString(key));
    }

    private void setDouble(MaterialEditText editText, Bundle args, String key) {
        if (editText == null || args == null) {
            return;
        }
        editText.setText(String.valueOf(args.getDouble(key)));
    }

    private String pointName(final MaterialEditText prefix,
                             final MaterialEditText numberCome,
                             final MaterialEditText suffix) {
        String number = string(numberCome);
        if (number.isEmpty()) {
            return string(prefix) + string(suffix);
        }
        long value = 0;
        try {
            value = Long.parseLong(number);
        } catch (NumberFormatException ignored) {

        }
        String result = String.format(Locale.getDefault(), "%s%03d%s", string(prefix), value, string(suffix));
        return result.replace(",", "");
    }

    private boolean checkXYValidate(MaterialEditText x, MaterialEditText y) {
        return x != null && y != null &&
                x.getText() != null && y.getText() != null &&
                !x.getText().toString().isEmpty() && !y.getText().toString().isEmpty();
    }

    private static final String KEY_PROJECT_ID = "___project__id";
    private static final String KEY_POINT_TYPE = "___point_type__";
    private static final String KEY_POINT_NAME = "___point_name__";
    private static final String KEY_POINT_ID = "___point_id";
    private static final String KEY_POINT_X = "___point_x";
    private static final String KEY_POINT_Y = "___point_y";
    private static final String KEY_POINT_Z = "__point_z";

    private static final String KEY_TYPE = "_TYPE_";
    private static final String VALUE_NEW = "___new___";
    private static final String VALUE_EDIT = "___edit___";


    public static void start(FragmentManager fragmentManager,
                             String projectId,
                             String pointType) {
        Bundle args = new Bundle();
        args.putString(KEY_PROJECT_ID, projectId);
        args.putString(KEY_POINT_TYPE, pointType);
        show(fragmentManager, args);
    }


    public static void edit(FragmentManager fragmentManager,
                            String projectId,
                            String pointName,
                            String pointId,
                            String pointType,
                            double x,
                            double y,
                            double z) {
        Bundle args = new Bundle();
        args.putString(KEY_PROJECT_ID, projectId);
        args.putString(KEY_POINT_NAME, pointName);
        args.putString(KEY_POINT_ID, pointId);
        args.putString(KEY_POINT_TYPE, pointType);

        args.putDouble(KEY_POINT_X, x);
        args.putDouble(KEY_POINT_Y, y);
        args.putDouble(KEY_POINT_Z, z);
        args.putString(KEY_TYPE, VALUE_EDIT);

        show(fragmentManager, args);
    }


    private static void show(FragmentManager fragmentManager,
                             Bundle args) {
        PointFragment fragment = new PointFragment();
        if (args != null) {
            fragment.setArguments(args);
        }
        fragment.show(fragmentManager, PointFragment.class.getName());
    }

    public interface ICallback {
        void onGet(String pointName, double x, double y, double z);
    }

    private void showPointSameNameDialogHint() {
        ThreadPoolUtil.executeInMain(() -> {
            //点名重复，提示
            TipsFragment.tipsOnly(
                    getChildFragmentManager(),
                    getViewLifecycleOwner(),
                    requireContext(),
                    Tools.getString(R.string.same_project_ponit_point_cannot_repeate));
        });
    }

    private boolean isEdit() {
        if (getArguments() != null) {
            return VALUE_EDIT.equals(getArguments().getString(KEY_TYPE, VALUE_NEW));
        }
        return false;
    }

    private String getProjectId() {
        if (getArguments() != null) {
            return getArguments().getString(KEY_PROJECT_ID);
        }
        return null;
    }

    private String getPointId() {
        if (getArguments() != null) {
            return getArguments().getString(KEY_POINT_ID);
        }
        return null;
    }

    private String getPointType() {
        if (getArguments() != null) {
            return getArguments().getString(KEY_POINT_TYPE);
        }
        return null;
    }

    private boolean isKnownPoint(String pointType) {
        return PointRecord.TYPE_KNOWN_POINT.equals(pointType) || PointRecord.TYPE_GLOBAL_POINT.equals(pointType);
    }

    private void checkSave(ICallBackSave callBackSave) {
        String pointName = pointName(mBinding.prefix, mBinding.number, mBinding.suffix);//string(prefix) + string(number) + string(suffix);
        if (pointName.isEmpty()) {
            ToastUtils.showShort(R.string.point_name_cannot_be_empty);
            callBackSave.result(false);
            return;
        }

        if (isKnownPoint(getPointType()) && !checkXYValidate(mBinding.pointX, mBinding.pointY)) {
            ToastUtils.showShort(R.string.error_tips_known_point_x_y_cannt_be_empty);
            callBackSave.result(false);
            return;
        }

        ThreadPoolUtil.execute(() -> {
            boolean sameName = PointUtil.isSameName(getProjectId(), pointName, getPointId());
            if (sameName) {
                showPointSameNameDialogHint();
                callBackSave.result(false);
                return;
            }
            callBackSave.result(true);
        });
    }

    private void dismissPointFragment() {
        runOnUI(() -> {
            SoftKeyUtils.hideSoftKey();
            dismiss();
        });
    }

    private interface ICallBackSave {
        void result(boolean isCanSave);
    }


}
