package com.bingce.controlnetwork.fragment

import android.content.Context
import android.os.Bundle
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import androidx.annotation.MainThread
import androidx.fragment.app.DialogFragment
import blankj.utilcode.util.ToastUtils
import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory
import com.bingce.controlapphelper.util.CheckStorageUtil
import com.bingce.controlapphelper.util.StationUtil.getStationName
import com.bingce.controlapphelper.util.StationUtil.getStationPointName
import com.bingce.controlapphelper.util.StationUtil.isStationCanExportForHeightTraverse
import com.bingce.controlapphelper.util.Tools
import com.bingce.controlnetwork.BuildConfig
import com.bingce.controlnetwork.R
import com.bingce.controlnetwork.activity.fragmentcontainer.FragmentContainerActivity
import com.bingce.controlnetwork.databinding.FragmentSelectExportSurveyorLayoutBinding
import com.bingce.controlnetwork.fragment.base.BaseFragment
import com.bingce.controlnetwork.surveyor.method.ExportUtilsPla
import com.bingce.controlnetwork.ui.widget.treeview.TreeNode
import com.bingce.controlnetwork.ui.widget.treeview.TreeView
import com.bingce.controlnetwork.ui.widget.treeview.custom.MyNodeViewFactory
import com.bingce.file.FileOperator
import com.bingce.utils.ThreadPoolUtil
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.MainScope
import kotlinx.coroutines.cancel
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import java.io.File


class SelectExportHeightTraverseFragment : BaseFragment() {

    private var _binding: FragmentSelectExportSurveyorLayoutBinding? = null
    val mBinding get() = _binding!!

    private val mainScope = MainScope()

    val root: TreeNode = TreeNode.root()

    private val treeView: TreeView by lazy {
        TreeView(root, requireContext(), MyNodeViewFactory())
    }

    override fun onCreateView(
        inflater: LayoutInflater,
        container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View {
        _binding = FragmentSelectExportSurveyorLayoutBinding.inflate(inflater, container, false)
        return mBinding.root
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)
        mBinding.export.isEnabled = true

        mBinding.export.setOnClickListener {
            if (ExportUtilsPla.isCanExport()) {
                requestStoragePermission()
            }
        }

        initData()
    }

    override fun onDestroyView() {
        super.onDestroyView()
        _binding = null
        mainScope.cancel()
    }

    private fun initData() {
        mainScope.launch {
            withContext(Dispatchers.IO) {
                buildNodeTree()
            }
            treeView.setup(mBinding.list)
            mBinding.list.visibility = View.VISIBLE
            mBinding.loading.visibility = View.GONE
        }
    }

    private suspend fun buildNodeTree() {
        val projectDataSource = SurveyorDatabaseFactory.instance.projectDataSource
        val projectRecordList = projectDataSource.dataList

        for (projectRecord in projectRecordList) {
            //遍历项目
            //一级数据
            val projectNode =
                TreeNode(TreeNode.Value(projectRecord.projectName, projectRecord.id), 0)

            val routeList = SurveyorDatabaseFactory.instance.getRouteDataSource()
                .getByProjectId(projectRecord.getId())

            var projectHasData = false

            for (routeRecord in routeList) {
                //遍历路线
                //二级数据
                val periodNode =
                    TreeNode(TreeNode.Value(routeRecord.name, routeRecord.id), 1)

                val stationList =
                    SurveyorDatabaseFactory.instance.surveyorStation.getByProjectRoute(
                        projectRecord.getId(),
                        routeRecord.getId()
                    )

                var routeHasData = false

                for (stationRecord in stationList) {
                    //遍历测站
                    if (stationRecord == null) {
                        continue
                    }

                    val exportItem = ExportUtilsPla.ExportItem(
                        projectRecord.getProjectName(), projectRecord.getId(),
                        projectRecord.projectType,
                        routeRecord.getId(), routeRecord.name,
                        stationRecord.getId(),
                        getStationName(stationRecord),
                        getStationPointName(stationRecord)
                    )

                    //三级数据
                    val stationNode = TreeNode(
                        TreeNode.Value(exportItem, stationRecord.getId()),
                        2
                    )
                    //添加三级数据
                    periodNode.addChild(stationNode)

                    val stationCanExport = isStationCanExportForHeightTraverse(stationRecord)
                    stationNode.isEnable = stationCanExport
                    if (stationCanExport) {
                        projectHasData = true
                        routeHasData = true
                    }
                }

                //添加二级数据
                periodNode.isEnable = routeHasData
                projectNode.addChild(periodNode)

            }

            projectNode.isEnable = projectHasData
            //添加一级数据
            root.addChild(projectNode)
        }

    }

    companion object {
        @JvmField
        val TAG: String = SelectExportHeightTraverseFragment::class.java.name

        @JvmStatic
        fun newInstance(args: Bundle?): SelectExportHeightTraverseFragment {
            val fragment = SelectExportHeightTraverseFragment()
            fragment.arguments = args
            return fragment
        }

        @JvmStatic
        fun start(context: Context) {
            FragmentContainerActivity.start(
                context,
                TAG,
                R.string.export_height_traverse,
                null
            )
        }

    }


    private fun requestStoragePermission() {
        CheckStorageUtil.check(context) { exportHasPermission() }
    }

    private fun exportHasPermission() {
        val dialogFragment: DialogFragment = FullLoadingFragment.showLoading(childFragmentManager)
        doExport(root, requireContext(), object : ICallback {
            override fun onFailureInMainThread(msg: String?) {
                dialogFragment.dismiss()
                ToastUtils.showShort(R.string.tips_please_select_exported_data)
            }

            override fun onSuccessInMainThread(zip: File) {
                if (dialogFragment.isVisible) {
                    dialogFragment.dismiss()
                }
                TipsWith2OptionFragment.tipsCannotCancel(
                    childFragmentManager, viewLifecycleOwner,
                    getString(R.string.share),
                    getString(R.string.cancel),
                    getString(R.string.export_successful),
                    object : TipsWith2OptionFragment.SimpleCallback() {
                        override fun onConfirm() {
                            FileOperator.shareFile(context, zip, BuildConfig.APPLICATION_ID)
                        }
                    })
            }
        })
    }

    private interface ICallback {
        fun onFailureInMainThread(msg: String?)
        fun onSuccessInMainThread(zip: File)
    }

    @MainThread
    private fun doExport(
        root: TreeNode,
        context: Context,
        callback: ICallback
    ) {
        ThreadPoolUtil.execute {

            //层遍历node树，找到所有被选中的station
            val exportItem: MutableList<ExportUtilsPla.ExportItem> =
                ArrayList()
            val nodeList: MutableList<TreeNode> =
                ArrayList()
            nodeList.add(root)
            while (!nodeList.isEmpty()) {
                val first = nodeList.removeAt(0)
                if (first.isEnable && first.isSelected && first.value
                        .value is ExportUtilsPla.ExportItem
                ) {
                    exportItem.add(first.value.value as ExportUtilsPla.ExportItem)
                }
                nodeList.addAll(first.children)
            }
            if (exportItem.isEmpty()) {
                if (callback != null) {
                    ThreadPoolUtil.executeInMain {
                        callback.onFailureInMainThread(
                            context.getString(
                                R.string.error_tips_please_select_exported_station
                            )
                        )
                    }
                }
                return@execute
            }

            //执行导出
            val zip = ExportUtilsPla.doExportHeightTraverse(context, exportItem)
            //提示导出成功
            ThreadPoolUtil.executeInMain {
                if (zip == null) {
                    callback.onFailureInMainThread(Tools.getString(R.string.export_failure))
                } else {
                    callback.onSuccessInMainThread(zip)
                }
            }
        }
    }


}