package com.bingce.controlnetwork.fragment;

import android.content.Context;
import android.os.Bundle;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;

import androidx.annotation.MainThread;
import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.annotation.WorkerThread;
import androidx.fragment.app.DialogFragment;
import androidx.recyclerview.widget.RecyclerView;

import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory;
import com.bingce.controlapphelper.datasource.database.project.IProjectDataSource;
import com.bingce.controlapphelper.datasource.database.project.ProjectRecord;
import com.bingce.controlapphelper.datasource.database.surveyorcycle.ISurveyorPeriodDataSource;
import com.bingce.controlapphelper.datasource.database.surveyorcycle.SurveyorPeriodRecord;
import com.bingce.controlapphelper.datasource.database.surveyorstation.ISurveyorStationDataSource;
import com.bingce.controlapphelper.datasource.database.surveyorstation.SurveyorStationRecord;
import com.bingce.controlapphelper.util.CheckStorageUtil;
import com.bingce.controlapphelper.util.StationUtil;
import com.bingce.controlapphelper.util.Tools;
import com.bingce.controlnetwork.BuildConfig;
import com.bingce.controlnetwork.R;
import com.bingce.controlnetwork.activity.fragmentcontainer.FragmentContainerActivity;
import com.bingce.controlnetwork.fragment.base.BaseFragment;
import com.bingce.controlnetwork.newui.configfragment.SelectExportType;
import com.bingce.controlnetwork.surveyor.method.ExportUtilsPla;
import com.bingce.controlnetwork.ui.widget.treeview.TreeNode;
import com.bingce.controlnetwork.ui.widget.treeview.TreeView;
import com.bingce.controlnetwork.ui.widget.treeview.custom.MyNodeViewFactory;
import com.bingce.file.FileOperator;
import com.bingce.utils.ThreadPoolUtil;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import blankj.utilcode.util.ToastUtils;

/**
 * 观测数据导出（基于选择的站的点数据导出）
 */
public class SelectExportSurveyorRecordFragment extends BaseFragment {
    public static final String TAG = SelectExportSurveyorRecordFragment.class.getName();
    private TreeNode root;
    private TreeView treeView;
    private boolean isOpen = false;

    public static final String KEY_DATA = "key_data";
    private ArrayList<SelectExportType> selectExportTypeList;

    @Nullable
    @Override
    public View onCreateView(@NonNull LayoutInflater inflater, @Nullable ViewGroup container, @Nullable Bundle savedInstanceState) {
        root = TreeNode.root();
        if (getContext() != null) {
            treeView = new TreeView(root, getContext(), new MyNodeViewFactory());
        }
        View rootView = inflater.inflate(R.layout.fragment_select_export_surveyor_layout, container, false);
        View exportBtn = rootView.findViewById(R.id.export);
        exportBtn.setEnabled(true);

        exportBtn.setOnClickListener(v -> {
            if (ExportUtilsPla.isCanExport()) {
                requestStoragePermission();
            }
        });

        ThreadPoolUtil.execute(() -> {
            buildNodeTree(root);
            ThreadPoolUtil.executeInMain(() -> {
                RecyclerView list = rootView.findViewById(R.id.list);
                treeView.setup(list);
                list.setVisibility(View.VISIBLE);
                View loading = rootView.findViewById(R.id.loading);
                loading.setVisibility(View.GONE);
            });
        });
        return rootView;
    }

    @Override
    public void onViewCreated(@NonNull View view, @Nullable Bundle savedInstanceState) {
        super.onViewCreated(view, savedInstanceState);
        selectExportTypeList = getSelectExportTypeList();
    }

    @MainThread
    private void doExport(TreeNode root,
                          Context context,
                          ICallback callback) {
        ThreadPoolUtil.execute(() -> {
            //层遍历node树，找到所有被选中的station
            List<ExportUtilsPla.ExportItem> exportItem = new ArrayList<>();
            List<TreeNode> nodeList = new ArrayList<>();
            nodeList.add(root);
            while (!nodeList.isEmpty()) {
                TreeNode first = nodeList.remove(0);
                if (first.isEnable() && first.isSelected() && first.getValue().getValue() instanceof ExportUtilsPla.ExportItem) {
                    exportItem.add((ExportUtilsPla.ExportItem) first.getValue().getValue());
                }
                nodeList.addAll(first.getChildren());
            }
            if (exportItem.isEmpty()) {
                if (callback != null) {
                    ThreadPoolUtil.executeInMain(() -> callback.onFailureInMainThread(context.getString(R.string.error_tips_please_select_exported_station)));
                }
                return;
            }

            //执行导出
            File zip = ExportUtilsPla.doExport(context, exportItem, selectExportTypeList);
            //提示导出成功
            if (callback != null) {
                ThreadPoolUtil.executeInMain(() -> {
                    if (zip == null) {
                        callback.onFailureInMainThread(Tools.getString(R.string.export_failure));
                    } else {
                        callback.onSuccessInMainThread(zip);
                    }
                });
            }
        });
    }

    private interface ICallback {
        void onFailureInMainThread(String msg);

        void onSuccessInMainThread(File zip);
    }

    /**
     * 根据
     * <p>
     * （项目->测期->测站）
     * <p>
     * 三级结构，构建node树
     * <p>
     * 检测每个测站数据是否正常记录，如果没有则将其设置为不可选中状态
     *
     * @param root node树的根节点
     */
    @WorkerThread
    private static void buildNodeTree(TreeNode root) {
        IProjectDataSource projectDataSource = SurveyorDatabaseFactory.instance.getProjectDataSource();
        List<ProjectRecord> projectRecordList = projectDataSource.getDataList();
        for (ProjectRecord projectRecord : projectRecordList) {//遍历项目
            TreeNode projectNode =
                    new TreeNode(new TreeNode.Value(projectRecord.projectName, projectRecord.id), 0);
            ISurveyorPeriodDataSource surveyorPeriodDataSource = SurveyorDatabaseFactory.instance.getSurveyorPeriod();
            List<SurveyorPeriodRecord> surveyorPeriodRecordList = surveyorPeriodDataSource.getDataListSyncByProjectId(projectRecord.getId());

            boolean projectHasData = false;
            for (SurveyorPeriodRecord periodRecord : surveyorPeriodRecordList) {//遍历测期
                TreeNode periodNode = new TreeNode(new TreeNode.Value(periodRecord.getName(), periodRecord.id), 1);
                ISurveyorStationDataSource surveyorStationDataSource = SurveyorDatabaseFactory.instance.getSurveyorStation();
                List<SurveyorStationRecord> stationRecordList = surveyorStationDataSource.getDataList(projectRecord.getId(), periodRecord.getId());

                boolean periodHasData = false;
                for (SurveyorStationRecord stationRecord : stationRecordList) {//遍历测站
                    if (stationRecord == null) {
                        continue;
                    }
                    ExportUtilsPla.ExportItem exportItem =
                            new ExportUtilsPla.ExportItem(projectRecord.getProjectName(), projectRecord.getId(),
                                    projectRecord.projectType,
                                    periodRecord.getId(), periodRecord.getName(),
                                    stationRecord.getId(),
                                    StationUtil.getStationName(stationRecord),
                                    StationUtil.getStationPointName(stationRecord));
                    TreeNode stationNode = new TreeNode(new TreeNode.Value(exportItem, stationRecord.getId()),
                            2);
                    periodNode.addChild(stationNode);

                    boolean stationCanExport = StationUtil.isStationCanExport(stationRecord);
                    stationNode.setEnable(stationCanExport);
                    if (stationCanExport) {
                        projectHasData = true;
                        periodHasData = true;
                    }
                }

                periodNode.setEnable(periodHasData);
                projectNode.addChild(periodNode);
            }

            projectNode.setEnable(projectHasData);
            root.addChild(projectNode);
        }
    }

    public static SelectExportSurveyorRecordFragment newInstance(Bundle args) {
        SelectExportSurveyorRecordFragment fragment = new SelectExportSurveyorRecordFragment();
        fragment.setArguments(args);
        return fragment;
    }

    public static void start(Context context, Bundle bundle) {
        FragmentContainerActivity.start(context, TAG, R.string.export, bundle);
    }

    public void requestStoragePermission() {
        CheckStorageUtil.check(getContext(), this::exportHasPermission);
    }

    private void exportHasPermission() {
        DialogFragment dialogFragment = FullLoadingFragment.showLoading(getChildFragmentManager());
        doExport(root, getContext(), new ICallback() {
            @Override
            public void onFailureInMainThread(String msg) {
                dialogFragment.dismiss();
                ToastUtils.showShort(R.string.tips_please_select_exported_data);
            }

            @Override
            public void onSuccessInMainThread(File zip) {
                if (dialogFragment.isVisible()) {
                    dialogFragment.dismiss();
                }
                TipsWith2OptionFragment.tipsCannotCancel(getChildFragmentManager(), getViewLifecycleOwner(),
                        getString(R.string.share),
                        getString(R.string.cancel),
                        getString(R.string.export_successful),
                        new TipsWith2OptionFragment.SimpleCallback() {
                            @Override
                            public void onConfirm() {
                                FileOperator.shareFile(getContext(), zip, BuildConfig.APPLICATION_ID);
                            }
                        });
            }
        });
    }

    private ArrayList<SelectExportType> getSelectExportTypeList() {
        Bundle arguments = getArguments();
        if (arguments != null) {
            return arguments.getParcelableArrayList(KEY_DATA);
        }
        return null;
    }

}