package com.bingce.controlnetwork.fragment.projectedit

import android.content.Context
import android.content.Intent
import android.os.Bundle
import android.text.TextUtils
import android.view.View
import androidx.activity.result.ActivityResultLauncher
import androidx.fragment.app.DialogFragment
import androidx.fragment.app.Fragment
import androidx.fragment.app.FragmentManager
import androidx.room.withTransaction
import blankj.utilcode.util.ToastUtils
import com.bingce.controlapphelper.datasource.database.ProjectType
import com.bingce.controlapphelper.datasource.database.SurveyorDataBase
import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory
import com.bingce.controlapphelper.datasource.database.config.ConfigConstants
import com.bingce.controlapphelper.datasource.database.project.ProjectRecord
import com.bingce.controlapphelper.datasource.database.surveyorcycle.SurveyorPeriodRecord
import com.bingce.controlapphelper.datasource.database.tolerance.load.ToleranceLoadUtil
import com.bingce.controlnetwork.R
import com.bingce.controlnetwork.fragment.AbstractFullScreenDialogFragment
import com.bingce.controlnetwork.fragment.ToleranceSelectFragment
import com.bingce.controlnetwork.ui.widget.dialog.LoadingUtilKt
import com.bingce.controlnetwork.util.DefaultNameUtil
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.MainScope
import kotlinx.coroutines.cancel
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext

abstract class AbstractProjectEditFragment : AbstractFullScreenDialogFragment() {
    companion object {

        const val KEY_PROJECT_ID = "___project__id"
        const val KEY_PROJECT_TYPE = "___project__type"
        const val KEY_PROJECT_NAME = "___project___name___"
        const val KEY_TOLERANCE_ID = "___tolerance_id"

        @JvmStatic
        fun editProject(
            fragmentManager: FragmentManager,
            context: Context,
            projectId: String, projectType: String?, toleranceId: String?, projectName: String?
        ) {
            newInstance(
                context, projectId, projectType,
                toleranceId, projectName
            ).show(fragmentManager, "_edit_project_fragment")
        }

        @JvmStatic
        fun createProject(
            fragmentManager: FragmentManager,
            context: Context,
            projectType: String?
        ) {
            newInstance(context, null, projectType, null, null)
                .show(fragmentManager, "__create_project_fragment__")
        }

        private fun newInstance(
            context: Context, projectId: String?, projectType: String?,
            toleranceId: String?, projectName: String?
        ): DialogFragment {
            val fragment =
                if (projectType == ProjectType.HEIGHT_TRAVERSE.flag) ProjectEditHeightTraverseFragment()
                else ProjectEditTriangleFragment()

            val args = Bundle()
            if (projectId.isNullOrEmpty()) {
                setArgs(args, context.getString(R.string.title_create_new_project), null)
                args.putString(KEY_PROJECT_TYPE, projectType)
            } else {
                setArgs(args, context.getString(R.string.title_edit_project), null)
                args.putString(KEY_PROJECT_ID, projectId)
                args.putString(KEY_PROJECT_TYPE, projectType)
                args.putString(KEY_TOLERANCE_ID, toleranceId)
                args.putString(KEY_PROJECT_NAME, projectName)
            }
            fragment.arguments = args
            return fragment
        }

    }


    protected val scope = MainScope()

    private lateinit var selectedToleranceId: String

    private lateinit var selectToleranceLauncher: ActivityResultLauncher<Intent>


    protected fun initSelectToleranceLauncher(fragment: Fragment) {
        selectToleranceLauncher = ToleranceSelectFragment.createActivityLauncher(
            fragment
        ) {
            updateToleranceData(it)
        }
    }


    fun getProjectId() = arguments?.getString(KEY_PROJECT_ID)
    fun getProjectType() = arguments?.getString(KEY_PROJECT_TYPE)
    fun getProjectName() = arguments?.getString(KEY_PROJECT_NAME)
    fun getToleranceId() = arguments?.getString(KEY_TOLERANCE_ID)

    override fun onDestroyView() {
        super.onDestroyView()
        scope.cancel()
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)
        initView()
        initData()
    }

    protected open fun initView() {
        getToleranceBtn().setOnClickListener {
            selectTolerance()
        }
    }

    protected abstract fun getCurrentProjectName(): String
    abstract fun getShowSlopSwitch(): Boolean
    abstract fun setToleranceName(name: String)
    abstract fun setProjectName(name: String)
    abstract fun setShowSlopSwitch(isChecked: Boolean)
    abstract fun getToleranceBtn(): View

    //存其他的数据，如果是高程导线存自己的配置信息，没有自己的数据就空实现即可
    abstract suspend fun saveOtherRecord(projectId: String)


    private fun updateToleranceData(toleranceId: String?) {
        toleranceId?.let {
            scope.launch {
                val toleranceDetailRecord = withContext(Dispatchers.IO) {
                    SurveyorDatabaseFactory.instance
                        .toleranceDetailDataSource
                        .getRecordSync(toleranceId)
                }
                toleranceDetailRecord?.let {
                    val tjToleranceLoad = ToleranceLoadUtil.getTriangleToleranceLoad()
                    val groupRecord = tjToleranceLoad.groupOf(it.groupId)
                    val levelRecord = tjToleranceLoad.levelOf(it.levelId)

                    val lineName = if (levelRecord?.name == null) "" else "-${levelRecord.name}"

                    setToleranceName("${groupRecord?.name ?: ""}${lineName}-${it.name}")

                    selectedToleranceId = toleranceId
                }
            }
        }
    }

    protected open fun initData() {
        initToleranceData()
        //设置项目名称
        setProjectName(getProjectName() ?: "")//DefaultNameUtil.getDefaultProjectName()
        updateShowSlopData(getProjectId())
    }

    /**
     * 初始化限差数据
     */
    private fun initToleranceData() {
        val projectType = getProjectType()

        val newDefaultTolerance = ToleranceLoadUtil.getTriangleToleranceLoad().getNewDefaultTolerance(projectType)
        val toleranceId =
            getToleranceId() ?: newDefaultTolerance?.id
        updateToleranceData(toleranceId)
    }

    private fun updateShowSlopData(projectId: String?) {
        if (projectId == null) {
            val checked = when (getProjectType()) {
                ProjectType.TRIANGLE.flag -> true
                ProjectType.CONNECTION.flag -> false
                else -> true
            }
            setShowSlopSwitch(checked)
        } else {
            scope.launch {
                val projectRecord = withContext(Dispatchers.IO) {
                    SurveyorDatabaseFactory.instance
                        .projectDataSource
                        .getRecord(projectId)
                }
                projectRecord?.let {
                    setShowSlopSwitch(it.isShowSlopDistance)
                }
            }
        }
    }

    private fun selectTolerance() {
        //选择限差
        ToleranceSelectFragment.start(
            context,
            getToleranceId(),
            getProjectType(),
            selectToleranceLauncher
        )
    }

    private fun checkBase(): Boolean {
        val projectName = getCurrentProjectName()
        if (TextUtils.isEmpty(projectName)) {
            ToastUtils.showShort(R.string.tips_need_select_input_project_name)
            return false
        }

        if (!::selectedToleranceId.isInitialized || TextUtils.isEmpty(selectedToleranceId)) {
            ToastUtils.showShort(R.string.tips_need_select_project_tolerance)
            return false
        }
        return true
    }

    override fun onConfirm() {
        if (!checkBase()) return
        saveRecord()
    }

    private fun saveRecord() {
        scope.launch {
            LoadingUtilKt.show(requireContext(), null)

            //开始事务
            SurveyorDataBase.getInstance().withTransaction {
                if (isEdit()) {
                    savaRecordEdit(getCurrentProjectName(), getShowSlopSwitch())
                } else {
                    val projectRecord = ProjectRecord(getCurrentProjectName())
                    projectRecord.projectType = getProjectType()
                    projectRecord.setToleranceId(selectedToleranceId)
                    projectRecord.isShowSlopDistance = getShowSlopSwitch()
                    savaRecordNew(projectRecord)
                }
            }

            LoadingUtilKt.dismiss()

            super.onConfirm()
        }
    }

    private suspend fun savaRecordNew(projectRecord: ProjectRecord) {
        //存储
        SurveyorDatabaseFactory.instance
            .projectDataSource
            .saveRecordSync(projectRecord)

        if (ProjectType.HEIGHT_TRAVERSE.flag != projectRecord.getProjectType()) {
            //非高程导线
            //创建默认测期
            val surveyorPeriodRecord =
                SurveyorPeriodRecord(projectRecord.getId(), DefaultNameUtil.getDefaultPeriodName())
            SurveyorDatabaseFactory.instance
                .surveyorPeriod
                .saveRecordSync(surveyorPeriodRecord)

            //创建新项目的时候存默认的测期
            SurveyorDatabaseFactory.instance.configDataSource.setSync(
                ConfigConstants.KEY_DEFAULT_PERIOD_ID,
                surveyorPeriodRecord.getId()
            )
        } else {
            //创建高程导线时候清空默认路线
            SurveyorDatabaseFactory.instance.configDataSource.setSync(
                ConfigConstants.KEY_DEFAULT_ROUTE_ID,
                null
            )
        }

        //创建新项目的时候存默认的项目，无论是否是高程导线
        SurveyorDatabaseFactory.instance.configDataSource.setSync(
            ConfigConstants.KEY_DEFAULT_PROJECT_ID,
            projectRecord.getId()
        )

        saveOtherRecord(projectRecord.getId())
    }

    private suspend fun savaRecordEdit(currentProjectName: String, showSlopSwitch: Boolean) {
        val projectRecord = SurveyorDatabaseFactory.instance
            .projectDataSource
            .getRecord(getProjectId())

        projectRecord.setProjectName(currentProjectName)
        projectRecord.setToleranceId(selectedToleranceId)
        projectRecord.isShowSlopDistance = showSlopSwitch

        //存储
        SurveyorDatabaseFactory.instance
            .projectDataSource
            .saveRecordSync(projectRecord)

        saveOtherRecord(getProjectId()!!)
    }

    protected fun isEdit() = getProjectId() != null


}