package com.bingce.controlnetwork.func.level.export

import android.text.TextUtils
import androidx.annotation.WorkerThread
import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory
import com.bingce.controlapphelper.datasource.database.leveling.LevelingDbFactory
import com.bingce.controlapphelper.datasource.database.leveling.stationsetting.LevelingStationSettingRecord
import com.bingce.controlapphelper.datasource.database.point.PointRecord
import com.bingce.controlapphelper.leveling.detaction.LevelFormatUtil
import com.bingce.controlapphelper.model.LevelingMeasureOrder
import com.bingce.controlapphelper.model._N
import com.bingce.controlapphelper.model._S
import com.bingce.controlapphelper.model._T
import com.bingce.controlapphelper.surveyor.method.ExportStationMethod
import com.bingce.controlapphelper.surveyor.method.model.MeasureConstant
import java.io.File

class LevelExportSingalHeight {
    private var routeIndex = 0

    private val singalHeightSb = StringBuilder()
    private val rulerLengthSb = StringBuilder()
    private val pointNumNameSb = StringBuilder()
    private val heightDistanceSb = StringBuilder()
    private val pointKnownSb = StringBuilder()


    @WorkerThread
    fun export(projectFolder: File, projectId: String) {
        collectRulerLength(projectId)
        collectPoint(projectId)
        collectSingalHeight(projectId)

        val prepare4Folder = initPrepare4Folder(projectFolder)

        saveData(projectFolder, singalHeightSb, "SingalHeight.txt")
        saveData(prepare4Folder, rulerLengthSb, "尺长.txt")
        saveData(prepare4Folder, pointNumNameSb, "点名点号.txt")
        saveData(prepare4Folder, heightDistanceSb, "高差文件.txt")
        saveData(prepare4Folder, pointKnownSb, "已知.txt")
    }

    private fun saveData(projectFolder: File, data: StringBuilder, fileName: String) {
        if (data.isNotEmpty()) {
            ExportStationMethod.save(
                data.toString(),
                projectFolder,
                fileName
            )
        }
    }

    private fun collectSingalHeight(projectId: String) {
        val routeList = LevelingDbFactory.instance.levelDataSource.getByProjectIdSync(projectId)

        for (route in routeList) {

            val routeId = route.getId()

            val pointNameStart = getPointName(route.startPointId)
            val pointNameEnd = getPointName(route.endPointId)

            val levelingMeasureOrder = route.levelingMeasureOrder
            val isHasTwiceStation = when (levelingMeasureOrder) {
                LevelingMeasureOrder.SIM_BF,
                LevelingMeasureOrder.SIM_BFFB -> {
                    true
                }

                else -> false
            }

            val routeLastStationSetting =
                LevelingDbFactory.instance.stationSettingDataSource.getLatest(routeId)

            appendData(
                projectId,
                routeId,
                pointNameStart,
                pointNameEnd,
                false,
                routeLastStationSetting
            )
            if (isHasTwiceStation) {
                appendData(
                    projectId,
                    routeId,
                    pointNameStart,
                    pointNameEnd,
                    true,
                    routeLastStationSetting
                )
            }
        }

        heightDistanceSb.append("END")
    }

    /**
     * 点名点号
     * 已知点
     */
    private fun collectPoint(projectId: String) {
        val pointList =
            SurveyorDatabaseFactory.instance.pointDataSource.findByProjectIdSync(projectId)
        for (point in pointList) {
            if (PointRecord.TYPE_KNOWN_POINT == point.pointType) {
                pointKnownSb.append(point.getName()).append(_S).append(point.getZFormat(4))
            }
            pointNumNameSb.append(point.getName()).append(_S).append("")
                .append(_N)
        }
        pointNumNameSb.append("END")
        pointKnownSb.append("END")
    }

    /**
     * 收集尺长
     */
    private fun collectRulerLength(projectId: String) {
        val routeList = LevelingDbFactory.instance.levelDataSource.getByProjectIdSync(projectId)
        for (route in routeList) {
            val latest = LevelingDbFactory.instance.stationSettingDataSource.getLatest(route.id)

            val rulerCorrectionFactor = if (!TextUtils.isEmpty(latest.rulerCorrectionFactor)) {
                "${latest.rulerCorrectionFactor.toDouble() / 1000}"
            } else {
                "0"
            }

            rulerLengthSb.append("1").append(_S).append(rulerCorrectionFactor)
                .append(_N)
        }
        rulerLengthSb.append("END")
    }

    private fun appendData(
        projectId: String,
        routeId: String,
        pointNameStart: String,
        pointNameEnd: String,
        twiceStation: Boolean,
        routeLatestSetting: LevelingStationSettingRecord
    ) {

        val stationFirst =
            LevelingDbFactory.instance.stationDataSource.getStationFirstTwiceStation(
                projectId,
                routeId,
                twiceStation
            )

        val stationLatest =
            LevelingDbFactory.instance.stationDataSource.getStationLatestTwiceStation(
                projectId,
                routeId,
                twiceStation
            )

        val measureStation =
            LevelingDbFactory.instance.measureStationDataSource.getRecordSync(stationLatest.getId())
                ?: return

        singalHeightSb.append(getRouteNum()).append(_T)//编号
            .append(pointNameStart).append(_T)//起点
            .append(pointNameEnd).append(_T)//终点
            .append(LevelFormatUtil.formatHeight(measureStation.heightDiffTotal)).append(_T)//高差
            .append(LevelFormatUtil.formatHeight(measureStation.distanceTotal)).append(_T)//距离
            .append("0.00000").append(_T)//温度改正
            .append(getExportRulerCorrectionFactor(routeLatestSetting)).append(_T)//标尺改正系数
            .append("1").append(_T)//结束标志
            .append(routeLatestSetting.soil ?: "水泥").append(_T)//土质
            .append(stationFirst.createDate.time).append(_T)//开始时间
            .append(stationLatest.createDate.time).append(_T)//结束时间
            .append(routeLatestSetting.temperature ?: MeasureConstant.DEFAULT_TEMPERATURE)
            .append(_T)//温度
            .append(routeLatestSetting.firstStationFrontFootNumber ?: "").append(_T)//标尺1
            .append(routeLatestSetting.firstStationBackFootNumber ?: "").append(_T)//标尺2
            .append(measureStation.stationNumForenoonTotal).append(_T)//上午测站数
            .append(measureStation.stationNumAfternoonTotal).append(_T)//下午测站数
            .append(_N)

        heightDistanceSb.append(pointNameStart).append(_S)
            .append(pointNameEnd).append(_S)
            .append(LevelFormatUtil.formatHeight(measureStation.heightDiffTotal)).append(_S)//高差
            .append(LevelFormatUtil.formatHeight(measureStation.distanceTotal)).append(_S)//距离
            .append("1").append(_N)//结束标志

        routeIndex++
    }

    private fun getExportRulerCorrectionFactor(routeLatestSetting: LevelingStationSettingRecord): String {
        return if (!TextUtils.isEmpty(routeLatestSetting.rulerCorrectionFactor)) {
            "${routeLatestSetting.rulerCorrectionFactor.toDouble() / 1000}"
        } else {
            "0"
        }
    }

    private fun getPointName(pointId: String): String {
        return SurveyorDatabaseFactory.instance.pointDataSource.findByIdSync(pointId)?.getName()
            ?: ""
    }

    private fun getRouteNum(): String {
        return String.format("%04d", routeIndex + 1)
    }

    /**
     *  初始化平差准备文件夹
     */
    private fun initPrepare4Folder(projectFolder: File): File {
        val exportFolder = File(projectFolder, "平差准备文件")
        if (!exportFolder.exists()) {
            exportFolder.mkdir()
        }
        return exportFolder
    }

}