package com.bingce.controlnetwork.func.level.util

import androidx.annotation.WorkerThread
import blankj.utilcode.util.ToastUtils
import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory
import com.bingce.controlapphelper.datasource.database.leveling.LevelingDbFactory
import com.bingce.controlapphelper.datasource.database.leveling.measure.station.LevelingMeasureStationRecord
import com.bingce.controlapphelper.datasource.database.leveling.station.LevelingStationRecord
import com.bingce.controlapphelper.model.LevelingMeasureOrder
import com.bingce.controlapphelper.model.StationOddEven
import com.bingce.controlapphelper.util.RegExUtil
import com.bingce.utils.DateUtils

object LevelStationUtilTj {
    @WorkerThread
    @JvmStatic
    fun createStation(
        projectId: String,
        routeId: String,
        backPointId: String,
        firstStation: Boolean = false,
        endStation: Boolean = false
    ) {
        val routeRecord = LevelingDbFactory.instance.levelDataSource.getByIdSync(routeId)
        routeRecord?.let {
            val levelingMeasureOrder = it.levelingMeasureOrder

            //新建路线的时候创建一个水准测站
            val frontPoint =
                LevelPointUtilTj.createFrontPointByBackPointId(
                    routeId,
                    backPointId,
                    firstStation,
                    endStation
                )

            val stationNamePrefix = "测站"
            val lastStationNameIndex = getLastStationNameIndex(projectId, routeId)

            //创建测站的奇偶站
            val createStationOdd = getCreateStationOdd(projectId, routeId)

            when (levelingMeasureOrder) {
                LevelingMeasureOrder.SIM_BF, LevelingMeasureOrder.SIM_BFFB -> {
                    val relationId = DateUtils.genRandomTimeId()
                    saveOneStationRecord(
                        projectId,
                        routeId,
                        "${stationNamePrefix}${lastStationNameIndex.plus(1)}_L",
                        backPointId,
                        frontPoint.getId(),
                        false,
                        createStationOdd,
                        relationId,
                        endStation = false
                    )
                    saveOneStationRecord(
                        projectId,
                        routeId,
                        "${stationNamePrefix}${lastStationNameIndex.plus(1)}_R",
                        backPointId,
                        frontPoint.getId(),
                        true,
                        createStationOdd,
                        relationId,
                        endStation = endStation
                    )
                }

                LevelingMeasureOrder.A_BF, LevelingMeasureOrder.A_BFFB, LevelingMeasureOrder.A_FBBF -> {
                    saveOneStationRecord(
                        projectId,
                        routeId,
                        "${stationNamePrefix}${lastStationNameIndex.plus(1)}",
                        backPointId,
                        frontPoint.getId(),
                        false,
                        createStationOdd,
                        endStation = endStation
                    )
                }

                else -> {
                    saveOneStationRecord(
                        projectId,
                        routeId,
                        "${stationNamePrefix}${lastStationNameIndex.plus(1)}",
                        backPointId,
                        frontPoint.getId(),
                        false,
                        createStationOdd,
                        endStation = endStation
                    )
                }
            }
        }
    }

    @WorkerThread
    private fun getLastStationNameIndex(projectId: String, routeId: String): Int {
        val stationLatest =
            LevelingDbFactory.instance.stationDataSource.getStationLatest(projectId, routeId)
        return stationLatest?.stationName?.let {
            val findNum = RegExUtil.findNum(it)
            findNum?.toInt() ?: 0
        } ?: 0
    }

    @WorkerThread
    private fun getLastStationId(projectId: String, routeId: String): String? {
        return getStationLatest(projectId, routeId)?.id
    }

    private fun getStationLatest(projectId: String, routeId: String) =
        LevelingDbFactory.instance.stationDataSource.getStationLatest(projectId, routeId)

    @WorkerThread
    private fun saveOneStationRecord(
        projectId: String,
        routeId: String,
        stationName: String,
        backPointId: String,
        frontPointId: String,
        twiceStation: Boolean,
        oddStation: StationOddEven,
        relationId: String? = null,
        endStation: Boolean = false
    ) {
        val lastStationId = getLastStationId(projectId, routeId)

        val station = LevelingStationRecord.createStation(
            projectId,
            routeId,
            stationName,
            lastStationId,
            twiceStation,
            backPointId,
            frontPointId,
            0,
            oddStation,
            endStation
        )
        station.relationStationId = relationId
        LevelingDbFactory.instance.stationDataSource.save(station)
    }

    private fun getCreateStationOdd(
        projectId: String,
        routeId: String
    ): StationOddEven {
        val stationLatest = getStationLatest(projectId, routeId)
            ?: return StationOddEven.ODD_STATION
        return if (stationLatest.isOddStation) {
            StationOddEven.EVEN_STATION
        } else {
            StationOddEven.ODD_STATION
        }
    }

    @JvmStatic
    fun checkStationForResult(stationId: String): Boolean {
        val stationRecord = LevelingDbFactory.instance.stationDataSource.getRecordSync(stationId)
        if (stationRecord.isFinish) {
            return true
        } else {
            if (stationRecord == null || !stationRecord.isFull) {
                ToastUtils.showShort("该测站数据不完整！")
            }
            if (!stationRecord.isFinish) {
                ToastUtils.showShort("该测站未通过检测！")
            }
            return false
        }
    }

    @JvmStatic
    fun getStationFinishCount(projectId: String, routeId: String): Int {
        return LevelingDbFactory.instance.stationDataSource.getCountFinish(projectId, routeId)
    }

    /**
     * 获取要测量的测站列表
     */
    fun getMeasureStationList(projectId: String, routeId: String): List<LevelingStationRecord> {
        val stationLatest =
            LevelingDbFactory.instance.stationDataSource.getStationLatest(projectId, routeId)
        val returnList = mutableListOf<LevelingStationRecord>()
        if (stationLatest != null) {
            if (stationLatest.twiceStation) {
                //单程双转点
                val statonList =
                    LevelingDbFactory.instance.stationDataSource.getStationListByRelationId(
                        projectId,
                        routeId,
                        stationLatest.relationStationId
                    )
                returnList.addAll(statonList)
            } else {
                returnList.add(stationLatest)
            }
        }
        return returnList
    }

    @WorkerThread
    fun getStationRouteResult(stationRecord: LevelingStationRecord): String {
        val routeName =
            LevelingDbFactory.instance.levelDataSource.getByIdSync(stationRecord.routeId)?.routeName ?: ""

        val sb = StringBuilder()

        if (stationRecord.twiceStation) {
            val relationStation = LevelingDbFactory.instance.stationDataSource
                .getRelationStation(stationRecord.relationStationId, false)
            val twiceStationMeasureRecord = getMeasureStationRecord(relationStation.getId())
            appendMeasureResult(sb, routeName, twiceStationMeasureRecord)
        }

        val measureStationRecord = getMeasureStationRecord(stationRecord.getId())
        appendMeasureResult(sb, routeName, measureStationRecord)

        return sb.toString()
    }

    private fun appendMeasureResult(
        sb: StringBuilder,
        routeName: String,
        measureStationRecord: LevelingMeasureStationRecord
    ) {
        sb.append("${routeName}的总高差${measureStationRecord.heightDiffTotalShow},总距离${measureStationRecord.distanceTotalShow}")
            .append("\n")
    }

    private fun getMeasureStationRecord(stationId: String) =
        LevelingDbFactory.instance.measureStationDataSource.getRecordSync(stationId)

}