package com.bingce.controlnetwork.func.level.viewmodel

import android.content.Context
import android.text.TextUtils
import androidx.annotation.WorkerThread
import androidx.lifecycle.ViewModel
import androidx.lifecycle.viewModelScope
import androidx.room.withTransaction
import blankj.utilcode.util.ToastUtils
import com.bingce.controlnetwork.func.level.fragment.LevelRouteCreateEditFragment
import com.bingce.controlnetwork.func.level.util.LevelStationUtilTj
import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory
import com.bingce.controlapphelper.datasource.database.config.ConfigOperate
import com.bingce.controlapphelper.datasource.database.leveling.LevelingDb
import com.bingce.controlapphelper.datasource.database.leveling.LevelingDbFactory
import com.bingce.controlapphelper.datasource.database.leveling.route.LevelRouteRecord
import com.bingce.controlapphelper.datasource.database.point.PointRecord
import com.bingce.controlapphelper.dialog.loading.LoadingUtil
import com.bingce.controlapphelper.model.LevelRouteSave
import com.bingce.controlapphelper.util.Tools
import com.bingce.controlnetwork.R
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext

class LevelCreateOrEditRouteVm : ViewModel() {

    fun showCreateOrEditDialog(
        context: Context,
        projectId: String,
        routeId: String?
    ) {
        val title = if (routeId == null)
            Tools.getString(R.string.create_segment_survey)
        else Tools.getString(R.string.edit_segment_survey)

        LevelRouteCreateEditFragment.start(context, projectId, routeId, title)
//        LevelRouteCreateEditDialog.start(projectId, routeId, childFragmentManager)
    }

    /**
     * 新建路线 存默认路线
     * 编辑路线
     */
    private fun startSaveRoute(
        levelRouteSave: LevelRouteSave,
        callBack: (Boolean) -> Unit
    ) {
        viewModelScope.launch(Dispatchers.IO) {
            LevelingDb.getInstance().withTransaction {
                try {
                    if (levelRouteSave.routeId != null) {
                        editRoute(levelRouteSave)
                    } else {
                        createRoute(levelRouteSave)
                    }

                    dismissLoading()
                    mainBack(callBack, true)
                } catch (e: Exception) {
                    e.printStackTrace()
                    dismissLoading()
                    mainBack(callBack, false)
                }
            }
        }
    }

    private suspend fun mainBack(callBack: (Boolean) -> Unit, isSuccess: Boolean) {
        withContext(Dispatchers.Main) {
            callBack(isSuccess)
        }
    }

    private suspend fun editRoute(levelRouteSave: LevelRouteSave) {
        val routeRecord =
            LevelingDbFactory.instance.levelDataSource.getById(levelRouteSave.routeId)
        routeRecord?.let {
            //编辑路线的时候先把路线下的所有测站都删除
            deleteRouteStation(it.getId())

            it.routeName = levelRouteSave.routeName
            it.startPointId = levelRouteSave.startPointId
            it.endPointId = levelRouteSave.endPointId
            it.observationDirection = levelRouteSave.observationDirection
            it.levelingMeasureOrder = levelRouteSave.levelingMeasureOrder

            //更新路线
            LevelingDbFactory.instance.levelDataSource.update(it)
            //创建测站
            createRouteStation(it.projectId, it.getId(), levelRouteSave.startPointId)
        }
    }

    /**
     * 删除路线下的所有测站
     */
    private fun deleteRouteStation(routeId: String) {
        LevelingDbFactory.instance.stationDataSource.deleteByRouteId(routeId)
    }

    @WorkerThread
    private suspend fun createRoute(
        levelRouteSave: LevelRouteSave
    ) {
        val routeRecord =
            LevelRouteRecord.createRouteRecord(
                levelRouteSave.projectId,
                levelRouteSave.routeName,
                levelRouteSave.startPointId,
                levelRouteSave.endPointId,
                levelRouteSave.observationDirection,
                levelRouteSave.levelingMeasureOrder
            )
        //存路线
        LevelingDbFactory.instance.levelDataSource.saveRecord(routeRecord)

        //创建测站
        createRouteStation(
            levelRouteSave.projectId,
            routeRecord.getId(),
            levelRouteSave.startPointId
        )

        ConfigOperate.changeDefaultProjectAndRoute(
            levelRouteSave.projectId,
            routeRecord.getId()
        )
    }

    private fun createRouteStation(projectId: String, routeId: String, startPointId: String) {
        //创建测站
        LevelStationUtilTj.createStation(projectId, routeId, startPointId, true)
    }

    fun startSave(context: Context, levelRouteSave: LevelRouteSave, callBack: (Boolean) -> Unit) {
        showLoading(context)
        viewModelScope.launch(Dispatchers.IO) {
            val canEdit = checkCanEdit(levelRouteSave.projectId, levelRouteSave.routeId)
            if (!canEdit) {
                dismissLoading()
                ToastUtils.showShort(R.string.route_has_data_not_edit_hint)
                return@launch
            }

            var startPoint =
                SurveyorDatabaseFactory.instance.getPointDataSource()
                    .findByNameSync(levelRouteSave.startPointName, levelRouteSave.projectId)
            if (startPoint == null) {
                startPoint =
                    createPoint(levelRouteSave.projectId, levelRouteSave.startPointName!!, "")
                //存数据库
                SurveyorDatabaseFactory.instance.getPointDataSource().saveSync(startPoint)
            }

            var endPoint =
                SurveyorDatabaseFactory.instance.getPointDataSource()
                    .findByNameSync(levelRouteSave.endPointName, levelRouteSave.projectId)
            if (endPoint == null) {
                endPoint =
                    createPoint(levelRouteSave.projectId, levelRouteSave.endPointName!!, "")
                //存数据库
                SurveyorDatabaseFactory.instance.getPointDataSource().saveSync(endPoint)
            }

            levelRouteSave.startPointId = startPoint.getId()
            levelRouteSave.endPointId = endPoint.getId()

            startSaveRoute(
                levelRouteSave,
                callBack
            )
        }
    }

    private fun showLoading(context: Context) {
        LoadingUtil.show(context, context.getString(R.string.loading))
    }

    private suspend fun dismissLoading() {
        withContext(Dispatchers.Main) {
            LoadingUtil.dismiss()
        }
    }

    /**
     * @return
     */
    @WorkerThread
    private fun checkCanEdit(projectId: String, routeIdParam: String?): Boolean {
        //路线有数据时候不能编辑
        val routeId = routeIdParam ?: return true
        val routeRecord = SurveyorDatabaseFactory.instance.getRouteDataSource().getByIdSync(routeId)
        if (routeRecord != null) {
            val stationList = LevelingDbFactory.instance.getStationDataSource()
                .getStationList(projectId, routeRecord.getId(), false)
            if (stationList != null && stationList.isNotEmpty()) {
                for (stationRecord in stationList) {
                    val recordCount = LevelingDbFactory.instance.getMeasureBackFrontDataSource()
                        .getCountByStationId(stationRecord.getId())
                    if (recordCount > 0) {
                        return false
                    }
                }
            }
        }
        return true
    }

    @WorkerThread
    private fun createPoint(projectId: String, pointName: String, height: String): PointRecord {
        return createPointRecord(projectId, pointName, height)
    }

    private fun createPointRecord(
        projectId: String,
        pointText: String,
        height: String
    ): PointRecord {
        val pointType = if (isKnownPoint(height)) {
            PointRecord.TYPE_KNOWN_POINT
        } else {
            PointRecord.TYPE_UNKNOWN_POINT
        }
        val h: Double = stringToDouble(height)
        return PointRecord.point(
            projectId,
            pointType,
            pointText,
            "",
            0.0,
            0.0,
            0.0,
            "",
            -1,
        )
    }

    private fun isKnownPoint(height: String): Boolean {
        val heightD: Double = stringToDouble(height)
        return heightD != 0.0
    }

    private fun stringToDouble(value: String): Double {
        return if (TextUtils.isEmpty(value)) {
            0.0
        } else value.toDouble()
    }

    fun getEditRouteData(routeId: String, callBack: (LevelRouteSave) -> Unit) {
        viewModelScope.launch(Dispatchers.IO) {
            val routeRecord = LevelingDbFactory.instance.levelDataSource.getByIdSync(routeId)
            routeRecord?.let {
                val result = LevelRouteSave()
                result.observationDirection = it.observationDirection
                result.routeName = it.routeName

                val startPoint =
                    SurveyorDatabaseFactory.instance.pointDataSource.findByIdSync(it.startPointId)
                val endPoint =
                    SurveyorDatabaseFactory.instance.pointDataSource.findByIdSync(it.endPointId)
                result.startPointName = startPoint?.name
                result.endPointName = endPoint?.name
                result.levelingMeasureOrder=it.levelingMeasureOrder
                withContext(Dispatchers.Main) {
                    callBack(result)
                }
            }
        }
    }

}