package com.bingce.controlnetwork.func.level.viewmodel

import android.widget.TextView
import androidx.annotation.WorkerThread
import androidx.lifecycle.MutableLiveData
import androidx.lifecycle.ViewModel
import androidx.lifecycle.viewModelScope
import androidx.room.withTransaction
import blankj.utilcode.util.ToastUtils
import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory
import com.bingce.controlapphelper.datasource.database.leveling.LevelingDb
import com.bingce.controlapphelper.datasource.database.leveling.LevelingDbFactory
import com.bingce.controlapphelper.datasource.database.leveling.measure.station.LevelingMeasureStationRecord
import com.bingce.controlapphelper.datasource.database.leveling.station.LevelingStationRecord
import com.bingce.controlapphelper.datasource.database.leveling.tolerance.LevelingToleranceDetail
import com.bingce.controlapphelper.datasource.database.point.PointRecord
import com.bingce.controlapphelper.leveling.detaction.DetectionController
import com.bingce.controlapphelper.leveling.detaction.IDetectionHint
import com.bingce.controlapphelper.leveling.detaction.INextStation
import com.bingce.controlapphelper.leveling.detaction.LevelReceiveData
import com.bingce.controlapphelper.leveling.detaction.LevelingProcess
import com.bingce.controlapphelper.leveling.detaction.surveyorcache.SurveyorCacheData
import com.bingce.controlapphelper.model.B1
import com.bingce.controlapphelper.model.B2
import com.bingce.controlapphelper.model.F1
import com.bingce.controlapphelper.model.F2
import com.bingce.controlapphelper.model.LevelingMeasureOrder
import com.bingce.controlapphelper.model.LevelingMeasureOrderUtil
import com.bingce.controlapphelper.model.StationOddEven
import com.bingce.controlapphelper.util.SpannableStringUtil
import com.bingce.controlapphelper.util.Tools
import com.bingce.controlnetwork.R
import com.bingce.controlnetwork.databinding.ItemLevelSuveyorSurveyorBinding
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext

class LevelSuveyorSubVm : ViewModel() {

    var projectId: String? = null//项目id

    val stationLiveData = MutableLiveData<LevelingStationRecord>()
    val surveyorCacheDataLiveData = MutableLiveData<SurveyorCacheData>()
    val measureStationLiveData = MutableLiveData<LevelingMeasureStationRecord?>()

    private var detectionHintI: IDetectionHint? = null
    private var nextStationI: INextStation? = null

    private var process: LevelingProcess? = null
    private var toleranceDetail: LevelingToleranceDetail? = null

    fun initData(
        stationId: String,
        detectionHintI: IDetectionHint,
        nextStationI: INextStation?
    ) {
        this.detectionHintI = detectionHintI
        this.nextStationI = nextStationI
        viewModelScope.launch(Dispatchers.IO) {
            val station = getStation(stationId)

            val project =
                SurveyorDatabaseFactory.instance.projectDataSource.getRecord(station.projectId)
            projectId = project.getId()
            toleranceDetail =
                LevelingDbFactory.instance.toleranceDetailDataSource.getRecordSync(project.toleranceId)

            refreshStationProcess(station)
        }
    }

    @WorkerThread
    private fun refreshStationProcess(station: LevelingStationRecord?) {
        if (toleranceDetail == null) {
            ToastUtils.showShort("限差异常")
            return
        }
        if (station == null) {
            ToastUtils.showShort("测站异常")
            return
        }

        if (station.levelingMeasureOrder == null) {
            setStationLevelOrder(station)
        }

        process = DetectionController.createProcess(
            viewModelScope,
            station,
            toleranceDetail!!,
            detectionHintI,
            nextStationI,
            stationLiveData,
            surveyorCacheDataLiveData,
            measureStationLiveData
        )
    }

    /**
     * 接受数据
     */
    fun recieveData(data: LevelReceiveData) {
        process?.receiveData(data)
    }

    fun saveMeasureFrontPointId(pointId: String, callBack: () -> Unit) {
        val station = stationLiveData.value
        if (station != null && !LevelingMeasureOrderUtil.isMeasureBack(station)) {
            viewModelScope.launch(Dispatchers.IO) {
                station.frontSightPointId = pointId
                LevelingDbFactory.instance.stationDataSource.update(station)
                withContext(Dispatchers.Main) {
                    callBack()
                }
            }
        }
    }

    fun saveMeasureFrontPoint(point: PointRecord, callBack: () -> Unit) {
        val station = stationLiveData.value
        if (station != null) {
            viewModelScope.launch(Dispatchers.IO) {
                LevelingDb.getInstance().withTransaction {
                    SurveyorDatabaseFactory.instance.pointDataSource.saveSync(point)
                    station.frontSightPointId = point.getId()
                    LevelingDbFactory.instance.stationDataSource.update(station)
                    withContext(Dispatchers.Main) {
                        callBack()
                    }
                }
            }
        }
    }

    fun getMeasurePointName(station: LevelingStationRecord, callBack: (String) -> Unit) {
        viewModelScope.launch(Dispatchers.IO) {
            val measurePointId = LevelingMeasureOrderUtil.getMeasurePointId(station)
            val name =
                SurveyorDatabaseFactory.instance.pointDataSource.findByIdSync(measurePointId)?.name
            withContext(Dispatchers.Main) {
                callBack(name ?: "")
            }
        }
    }

    /**
     * 回退
     */
    fun startRollBack() {
        process?.startRollBack()
    }

    fun reMeasureStation() {
        process?.dealReMeasureStation()
    }

    fun getZ(callBack: (Double) -> Unit) {
        stationLiveData.value?.let { stationRecord ->
            viewModelScope.launch(Dispatchers.IO) {
                val routeRecord =
                    LevelingDbFactory.instance.levelDataSource.getByIdSync(stationRecord.routeId)
                val startPointHeight =
                    SurveyorDatabaseFactory.instance.pointDataSource.findByIdSync(routeRecord?.startPointId)
                        ?.getZ()
                        ?: 0.0
                withContext(Dispatchers.Main) {
                    callBack(startPointHeight)
                }
            }
        }
    }

    fun getCurrentOrderText(stationRecord: LevelingStationRecord): CharSequence {
        val textByLevelingMeasureOrder =
            LevelingMeasureOrderUtil.getTextByLevelingMeasureOrder(stationRecord.levelingMeasureOrder)
        val text = textByLevelingMeasureOrder.replaceFirst("(Sim|a])".toRegex(), "")

        return SpannableStringUtil.getSelectedText(
            text,
            stationRecord.currentPosition
        )
    }

    fun getCurrentOrderArray(stationRecord: LevelingStationRecord): Array<CharSequence> =
        when (stationRecord.levelingMeasureOrder) {
            LevelingMeasureOrder.A_BF -> {
                getArrayString(stationRecord, "BF", "FB")
            }

            LevelingMeasureOrder.A_BFFB -> {
                getArrayString(stationRecord, "BFFB", "FBBF")
            }

            LevelingMeasureOrder.A_FBBF -> {
                getArrayString(stationRecord, "FBBF", "BFFB")
            }

            else -> {
                val text =
                    LevelingMeasureOrderUtil.getTextByLevelingMeasureOrder(stationRecord.levelingMeasureOrder)
                        .replaceFirst("(Sim|a])".toRegex(), "")
                getArrayString(stationRecord, text, text)
            }
        }

    private fun getArrayString(
        stationRecord: LevelingStationRecord,
        text1: String,
        text2: String
    ): Array<CharSequence> {
        return if (stationRecord.oddEvenStation == StationOddEven.ODD_STATION) {
            arrayOf(
                SpannableStringUtil.getSelectedText(text1, stationRecord.currentPosition),
                text2
            )
        } else {
            arrayOf(
                text1,
                SpannableStringUtil.getSelectedText(text2, stationRecord.currentPosition),
            )
        }
    }

    private fun getStation(stationId: String): LevelingStationRecord {
        val station = LevelingDbFactory.instance.stationDataSource.getRecordSync(stationId)
        setStationLevelOrder(station)
        return station
    }

    private fun setStationLevelOrder(station: LevelingStationRecord) {
        val routeRecord = LevelingDbFactory.instance.levelDataSource.getByIdSync(station.routeId)
        station.levelingMeasureOrder = routeRecord?.levelingMeasureOrder
    }

    fun isShowOddEvenStation(station: LevelingStationRecord) =
        when (station.levelingMeasureOrder) {
            LevelingMeasureOrder.A_BF,
            LevelingMeasureOrder.A_BFFB,
            LevelingMeasureOrder.A_FBBF -> true

            else -> false
        }

    fun setMeasureTextColor(
        station: LevelingStationRecord,
        ilB1: ItemLevelSuveyorSurveyorBinding,
        ilB2: ItemLevelSuveyorSurveyorBinding,
        ilF1: ItemLevelSuveyorSurveyorBinding,
        ilF2: ItemLevelSuveyorSurveyorBinding
    ) {
        val measureWhichPoint = LevelingMeasureOrderUtil.getMeasureWhichPoint(station)
        resetTextColor(ilB1.tvDistanceTitle, ilB1.tvHeightTitle, ilB1.tvDisTance, ilB1.tvHeight)
        resetTextColor(ilB2.tvDistanceTitle, ilB2.tvHeightTitle, ilB2.tvDisTance, ilB2.tvHeight)
        resetTextColor(ilF1.tvDistanceTitle, ilF1.tvHeightTitle, ilF1.tvDisTance, ilF1.tvHeight)
        resetTextColor(ilF2.tvDistanceTitle, ilF2.tvHeightTitle, ilF2.tvDisTance, ilF2.tvHeight)
        when (measureWhichPoint) {
            B1 -> {
                selectTextColor(
                    ilB1.tvDistanceTitle,
                    ilB1.tvHeightTitle,
                    ilB1.tvDisTance,
                    ilB1.tvHeight
                )
            }

            B2 -> {
                selectTextColor(
                    ilB2.tvDistanceTitle,
                    ilB2.tvHeightTitle,
                    ilB2.tvDisTance,
                    ilB2.tvHeight
                )
            }

            F1 -> {
                selectTextColor(
                    ilF1.tvDistanceTitle,
                    ilF1.tvHeightTitle,
                    ilF1.tvDisTance,
                    ilF1.tvHeight
                )
            }

            F2 -> {
                selectTextColor(
                    ilF2.tvDistanceTitle,
                    ilF2.tvHeightTitle,
                    ilF2.tvDisTance,
                    ilF2.tvHeight
                )
            }
        }

    }

    private fun resetTextColor(
        tvDisTanceTitle: TextView,
        tvHeightTitle: TextView,
        tvDisTance: TextView,
        tvHeight: TextView
    ) {
        tvDisTanceTitle.setTextColor(Tools.getColor(R.color.color_333333))
        tvHeightTitle.setTextColor(Tools.getColor(R.color.color_333333))
        tvDisTance.setTextColor(Tools.getColor(R.color.color_333333))
        tvHeight.setTextColor(Tools.getColor(R.color.color_333333))
    }

    private fun selectTextColor(
        tvDisTanceTitle: TextView,
        tvHeightTitle: TextView,
        tvDisTance: TextView,
        tvHeight: TextView
    ) {
        tvDisTanceTitle.setTextColor(Tools.getColor(R.color.value_1_selected))
        tvHeightTitle.setTextColor(Tools.getColor(R.color.value_1_selected))
        tvHeight.setTextColor(Tools.getColor(R.color.value_1_selected))
        tvDisTance.setTextColor(Tools.getColor(R.color.value_1_selected))
    }

    fun isEndStation(station: LevelingStationRecord, callBack: (Boolean) -> Unit) {
        viewModelScope.launch(Dispatchers.IO) {
            val routeRecord = LevelingDbFactory.instance.levelDataSource.getById(station.routeId)
            val endStation = routeRecord?.endPointId == station.frontSightPointId
            withContext(Dispatchers.Main) {
                callBack(endStation)
            }
        }
    }

    fun nextStation() {
        process?.nextStationCheck()
    }


}