package com.bingce.controlnetwork.func.level.viewmodel

import android.text.TextUtils
import androidx.annotation.WorkerThread
import androidx.lifecycle.ViewModel
import androidx.lifecycle.viewModelScope
import com.bingce.controlnetwork.func.level.util.LevelStationUtilTj
import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory
import com.bingce.controlapphelper.datasource.database.leveling.LevelingDbFactory
import com.bingce.controlapphelper.datasource.database.leveling.station.LevelingStationRecord
import com.bingce.controlapphelper.datasource.database.leveling.stationsetting.LevelingStationSettingRecord
import com.bingce.controlapphelper.leveling.detaction.ILevelRefreshBtn
import com.bingce.controlapphelper.leveling.detaction.INextStation
import com.bingce.controlapphelper.leveling.detaction.LevelCheckBtnUtil
import com.bingce.controlapphelper.model.LevelingMeasureOrder
import com.bingce.controlapphelper.model.LevelingStationState
import com.bingce.controlnetwork.func.level.util.LevelDateUtil
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext

class LevelSuveyorVm : ViewModel() {
    var routeId: String? = null
    private var levelRefreshBtnI: ILevelRefreshBtn? = null
    var currentFragmentPos = 0

    //测量测站列表
    var measureStationList = emptyList<LevelingStationRecord>()

    private var initVpCallBack: (() -> Unit)? = null
    private var switchPageCallBack: ((Int) -> Unit)? = null
    var nextStationI: INextStation? = null


    fun initDataStationList(projectId: String, routeId: String) {
        viewModelScope.launch(Dispatchers.IO) {
            getMeasureStationList(projectId, routeId)
        }
    }

    private suspend fun getMeasureStationList(projectId: String, routeId: String) {
        this.routeId = routeId
        val routeRecord = LevelingDbFactory.instance.levelDataSource.getByIdSync(routeId)
        routeRecord?.let {
            measureStationList = LevelStationUtilTj.getMeasureStationList(projectId, routeId)
            LevelCheckBtnUtil.reset(levelRefreshBtnI)
            //初始化按钮状态
            initBtnState(routeRecord.levelingMeasureOrder)
            withContext(Dispatchers.Main) {
                initVpCallBack?.invoke()
            }
        }
    }

    private fun initBtnState(levelingMeasureOrder: LevelingMeasureOrder) {
        for (station in measureStationList) {
            station.levelingMeasureOrder = levelingMeasureOrder
            refreshBtnActive(station)
        }
    }

    fun initData(
        levelRefreshBtnI: ILevelRefreshBtn,
        initVpCallBack: (() -> Unit),
        switchPageCallBack: ((Int) -> Unit),
        nextStationI: INextStation
    ) {
        this.levelRefreshBtnI = levelRefreshBtnI
        this.initVpCallBack = initVpCallBack
        this.switchPageCallBack = switchPageCallBack
        this.nextStationI = nextStationI
    }

    /**
     * 切换页面刷新按钮
     */
    fun switchRefreshBtnEnable() {
        levelRefreshBtnI?.refreshMeasureBtn()
        levelRefreshBtnI?.refreshNextStationBtn()
        levelRefreshBtnI?.refreshRollBackBtn()
        levelRefreshBtnI?.refreshReMeasureBtn()
    }

    private fun refreshBtnActive(station: LevelingStationRecord) {
        LevelCheckBtnUtil.refreshMeasureBtn(station)
        LevelCheckBtnUtil.refreshNextStationBtn(station)
        LevelCheckBtnUtil.refreshRollBackBtn(station)
        LevelCheckBtnUtil.refreshRemeasureStationBtn(station)
    }

    fun getCurrentStationId(): String? {
        if (measureStationList.size > currentFragmentPos) {
            return measureStationList[currentFragmentPos].getId()
        }
        return null
    }

    /**
     * 继续观测下一站
     */
    fun contineObserve(station: LevelingStationRecord) {
        viewModelScope.launch(Dispatchers.IO) {
            startNextStation(station, false)
        }
    }

    /**
     * 间歇
     */
    fun intermittent(station: LevelingStationRecord) {
        // TODO:  
    }

    /**
     * 结束测站
     */
    fun endStation(station: LevelingStationRecord) {
        viewModelScope.launch(Dispatchers.IO) {
            val checkStaionIsFinish = checkStaionIsFinish(station)
            if (!checkStaionIsFinish) {
                return@launch
            }

            station.endStation = true
            LevelingDbFactory.instance.stationDataSource.update(station)

            refreshBtnActive(station)
            switchRefreshBtnEnable()
        }
    }

    private suspend fun startNextStation(station: LevelingStationRecord, endStation: Boolean) {
        //单程双测 判断测站是否测完
        val checkStaionIsFinish = checkStaionIsFinish(station)
        if (!checkStaionIsFinish) {
            return
        }

        LevelStationUtilTj.createStation(
            station.getProjectId(),
            station.routeId,
            station.frontSightPointId,
            endStation = endStation
        )
        //重新获取下一站
        getMeasureStationList(station.getProjectId(), station.routeId)
    }

    /**
     * 回退测站
     */
    fun rollBackStation() {
        viewModelScope.launch(Dispatchers.IO) {
            var projectId = ""
            var routeId = ""
            for (station in measureStationList) {
                if (TextUtils.isEmpty(projectId)) {
                    projectId = station.projectId
                }
                if (TextUtils.isEmpty(routeId)) {
                    routeId = station.routeId
                }

                LevelingDbFactory.instance.stationDataSource.deleteById(station.getId())
                LevelingDbFactory.instance.measureStationDataSource.deleteById(station.getId())
                LevelingDbFactory.instance.measureBackFrontDataSource.deleteByStationId(station.getId())
            }
            getMeasureStationList(projectId, routeId)
        }
    }

    private fun checkStaionIsFinish(station: LevelingStationRecord): Boolean {
        //单程双测 判断测站是否测完
        var index = -1
        if (station.relationStationId != null) {
            val stationListByRelationId =
                LevelingDbFactory.instance.stationDataSource.getStationListByRelationId(
                    station.getProjectId(),
                    station.routeId,
                    station.relationStationId
                )
            for ((i, item) in stationListByRelationId.withIndex()) {
                if (item.stationState != LevelingStationState.FINISH) {
                    index = i
                    break
                }
            }
        } else {
            if (station.stationState != LevelingStationState.FINISH) {
                index = 0
            }
        }

        if (index != -1) {
            switchPageCallBack?.invoke(index)
            return false
        }
        return true
    }

    fun getCurrentStation(): LevelingStationRecord? {
        if (measureStationList.size > currentFragmentPos) {
            return measureStationList[currentFragmentPos]
        }
        return null
    }

    fun calLevelRouteResult(callBack: (String) -> Unit) {
        getCurrentStation()?.let {
            viewModelScope.launch(Dispatchers.IO) {
                val success = LevelStationUtilTj.checkStationForResult(it.getId())
                if (!success) {
                    return@launch
                }
                val stationRouteResult = LevelStationUtilTj.getStationRouteResult(it)
                withContext(Dispatchers.Main) {
                    callBack(stationRouteResult)
                }
            }
        }
    }

    fun getGetDadaAuto(routeId: String, callBack: (Boolean) -> Unit) {
        viewModelScope.launch(Dispatchers.IO) {
            val record = LevelingDbFactory.instance.stationSettingDataSource.getLatest(routeId)
            withContext(Dispatchers.Main) {
                if (record == null) {
                    callBack(true)
                } else {
                    callBack(!record.getDataManually)
                }
            }
        }
    }

    fun setGetDataAuto(routeId: String, auto: Boolean) {
        viewModelScope.launch(Dispatchers.IO) {
            var record = LevelingDbFactory.instance.stationSettingDataSource.getLatest(routeId)
            if (record == null) {
                record = LevelingStationSettingRecord()
                record.routeId = routeId
                record.getDataManually = !auto
                LevelingDbFactory.instance.stationSettingDataSource.save(record)
            } else {
                record.getDataManually = !auto
                LevelingDbFactory.instance.stationSettingDataSource.update(record)
            }
        }
    }

    /**
     * 是否设置过最后一站的测站设置
     */
    @WorkerThread
    fun isHaveSetEndStationSetting(stationId: String): Boolean {
        val stationSetting =
            LevelingDbFactory.instance.stationSettingDataSource.getByStationId(stationId)
        return stationSetting?.setEndStationSetting ?: false
    }

    fun checkSunTime(projectId: String, showDialog: (String?) -> Unit) {
        viewModelScope.launch(Dispatchers.IO) {
            val projectRecord = SurveyorDatabaseFactory.instance.projectDataSource.getRecord(projectId)
            val toleranceDetail =
                LevelingDbFactory.instance.toleranceDetailDataSource.getRecordSync(projectRecord.toleranceId)

            var hint: String? = null
            if (toleranceDetail.isCheckSunTime) {
                hint = LevelDateUtil.getLevelOperateSunTimeHint(
                    projectRecord.latitude.toDouble(),
                    projectRecord.longitude.toDouble()
                )
            }
            withContext(Dispatchers.Main) {
                showDialog(hint)
            }
        }
    }

}