package com.bingce.controlnetwork.http

import android.util.Log
import blankj.utilcode.util.ToastUtils
import com.bingce.controlapphelper.util.GsonUtil
import com.bingce.controlnetwork.BuildConfig
import com.google.gson.JsonParseException
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.withContext
import okhttp3.Interceptor
import okhttp3.MediaType.Companion.toMediaTypeOrNull
import okhttp3.OkHttpClient
import okhttp3.RequestBody
import okhttp3.RequestBody.Companion.toRequestBody
import okhttp3.logging.HttpLoggingInterceptor
import org.apache.http.conn.ConnectTimeoutException
import org.json.JSONException
import retrofit2.HttpException
import retrofit2.Retrofit
import retrofit2.converter.gson.GsonConverterFactory
import java.io.IOException
import java.net.SocketTimeoutException
import java.net.UnknownHostException
import java.util.concurrent.TimeUnit

object RetrofitClient {
    //根路径
    private const val BASE_URL = "https://www.wanandroid.com/"

    private val mMediaTypeJson = "application/json; charset=utf-8".toMediaTypeOrNull()

    private val retrofit = Retrofit.Builder()
        .baseUrl(BASE_URL)
        .addConverterFactory(GsonConverterFactory.create())
        .client(initOkhttpClient())
        .build()


    private fun initOkhttpClient(): OkHttpClient {
        return OkHttpClient.Builder()
            .retryOnConnectionFailure(true) //重试
            .connectTimeout(20, TimeUnit.SECONDS)
            .readTimeout(20, TimeUnit.SECONDS)
            .writeTimeout(20, TimeUnit.SECONDS)
//            .addNetworkInterceptor(initTokenInterceptor())
            .addInterceptor(initLogInterceptor())
            .build()
    }

    private fun initTokenInterceptor(): Interceptor = Interceptor { chain -> // 为请求添加 Token
        val token: String? = null
        val originalRequest = chain.request()
        if (token.isNullOrEmpty()) {
            chain.proceed(originalRequest)
        } else {
            val updateRequest = originalRequest.newBuilder().header("token", token).build()
            chain.proceed(updateRequest)
        }
    }


    /*
   * 日志拦截器
   * */
    private fun initLogInterceptor(): HttpLoggingInterceptor {
        if (!BuildConfig.DEBUG) {   //生产环境不打日志
            return HttpLoggingInterceptor()
        }

        val interceptor = HttpLoggingInterceptor { message -> Log.d("OkhttpClient", message) }
        interceptor.level = HttpLoggingInterceptor.Level.BODY
        return interceptor
    }

    // 请求数据转 Json 形式;
    open fun getJsonBody(data: Any): RequestBody {
        val strEntity = GsonUtil.toJson(data)
        return strEntity.toRequestBody(mMediaTypeJson)
    }

    fun <T> create(serviceClass: Class<T>): T = retrofit.create(serviceClass)

    inline fun <reified T> createService(): T = create(T::class.java)


    suspend inline fun <T> apiCall(crossinline call: suspend CoroutineScope.() -> T): T? {
        return withContext(Dispatchers.IO) {
            var res: T? = null
            try {
                res = call()
            } catch (e: Throwable) {
                Log.e("ApiCaller", "request error", e)
                // 请求出错，将状态码和消息封装为 ResponseResult
//                return@withContext ApiException.build(e).toResponse<T>()
                ToastUtils.showShort(ApiException.build(e).toString())
            }
            return@withContext res
        }
    }

    // 网络、数据解析错误处理
    class ApiException(val code: Int, override val message: String?, override val cause: Throwable? = null) :
        RuntimeException(message, cause) {
        companion object {
            // 网络状态码
            const val CODE_NET_ERROR = 4000
            const val CODE_TIMEOUT = 4080
            const val CODE_JSON_PARSE_ERROR = 4010
            const val CODE_SERVER_ERROR = 5000

            // 业务状态码
            const val CODE_AUTH_INVALID = 401

            fun build(e: Throwable): ApiException {
                return if (e is HttpException) {
                    ApiException(CODE_NET_ERROR, "网络异常(${e.code()},${e.message()})")
                } else if (e is UnknownHostException) {
                    ApiException(CODE_NET_ERROR, "网络连接失败，请检查后再试")
                } else if (e is ConnectTimeoutException || e is SocketTimeoutException) {
                    ApiException(CODE_TIMEOUT, "请求超时，请稍后再试")
                } else if (e is IOException) {
                    ApiException(CODE_NET_ERROR, "网络异常(${e.message})")
                } else if (e is JsonParseException || e is JSONException) {
                    // Json解析失败
                    ApiException(CODE_JSON_PARSE_ERROR, "数据解析错误，请稍后再试")
                } else {
                    ApiException(CODE_SERVER_ERROR, "系统错误(${e.message})")
                }
            }
        }

//        fun <T> toResponse(): ResponseResult<T> {
//            return ResponseResult(code, message)
//        }

        override
        fun toString(): String {
            return "code:$code,message:$message"
        }
    }
}
