package com.bingce.controlnetwork.newui.createeditpoint

import android.content.Context
import android.content.Intent
import android.text.TextUtils
import android.view.View
import androidx.annotation.StringRes
import androidx.annotation.WorkerThread
import androidx.lifecycle.lifecycleScope
import blankj.utilcode.util.ToastUtils
import com.bingce.controlapphelper.datasource.database.ProjectType
import com.bingce.controlapphelper.datasource.database.ProjectTypeValue
import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory
import com.bingce.controlapphelper.datasource.database.point.PointRecord
import com.bingce.controlapphelper.model.BundleConstants
import com.bingce.controlnetwork.R
import com.bingce.controlnetwork.databinding.ActivityNewUiCreateEditPointBinding
import com.bingce.controlnetwork.fragment.TipsFragment
import com.bingce.controlnetwork.newui.base.BaseBindingActivity
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext

/**
 * 新建或编辑点
 */
class CreateEditPointActvity : BaseBindingActivity<ActivityNewUiCreateEditPointBinding>() {

    private var editPoint: PointRecord? = null

    private var isKnown = false

    companion object {
        const val KEY_POINT_TYPE = "KEY_POINT_TYPE"
        const val KEY_EDIT_POINT = "KEY_EDIT_POINT"

        @JvmStatic
        fun launch(
            context: Context,
            projectId: String,
            projectType: String,
            pointType: String,
            eidtPoint: PointRecord? = null
        ) {
            context.startActivity(Intent(context, CreateEditPointActvity::class.java).apply {
                putExtra(BundleConstants.KEY_PROJECT_ID, projectId)
                putExtra(BundleConstants.KEY_PROJECT_TYPE, projectType)
                putExtra(KEY_POINT_TYPE, pointType)
                putExtra(KEY_EDIT_POINT, eidtPoint)
            })
        }
    }

    override fun getBinding(): ActivityNewUiCreateEditPointBinding {
        return ActivityNewUiCreateEditPointBinding.inflate(layoutInflater)
    }

    override fun initView() {
        mBinding.ilPointNum.tvTitle.setText(R.string.point_name)
        mBinding.ilPointNum.editText.setHint(R.string.please_enter_roll_call)

        setBaseUi()
        initListener()
    }

    private fun initIntentData() {
        val pointType = getPointType()
        isKnown =
            PointRecord.TYPE_KNOWN_POINT == pointType || PointRecord.TYPE_GLOBAL_POINT == pointType
        editPoint = getEditPoint()
    }

    private fun initListener() {
        mBinding.tvSaveContinue.setOnClickListener {
            startSaveContinue()
        }
        mBinding.tvSaveQuit.setOnClickListener {
            startSaveQuite()
        }
    }

    private fun startSaveQuite() {
        checkSave {
            if (editPoint == null) {
                saveRecordForCreate {
                    finish()
                }
            } else {
                saveRecordForEdit(editPoint!!) {
                    finish()
                }
            }
        }
    }

    private fun startSaveContinue() {
        checkSave {
            saveRecordForCreate()
            clearContent()
            val nextPointNum = getNextPointNum()
            mBinding.ilPointNum.editText.setText(nextPointNum)
        }
    }

    private fun checkSave(success: () -> Unit) {
        val pointNum = getPointNumber()
        if (TextUtils.isEmpty(pointNum)) {
            ToastUtils.showShort(R.string.please_enter_roll_call)
            return
        }

        if (isKnown) {
            if (mBinding.ilPointX.root.visibility == View.VISIBLE) {
                val x = getX()
                if (TextUtils.isEmpty(x)) {
                    val hint = R.string.please_input_x
//                    val hint =
//                        if (isPlatCood) R.string.please_input_x else R.string.please_input_longitude
                    ToastUtils.showShort(hint)
                    return
                }
            }
            if (mBinding.ilPointY.root.visibility == View.VISIBLE) {
                val y = getY()
                if (TextUtils.isEmpty(y)) {
                    val hint = R.string.please_input_y
//                    val hint =
//                        if (isPlatCood) R.string.please_input_y else R.string.please_input_latitude
                    ToastUtils.showShort(hint)
                    return
                }
            }
            val z = getZ()
            if (TextUtils.isEmpty(z)) {
                val hint = R.string.please_enter_elevation
//                val hint = if (isLevelProject()) {
//                    R.string.please_enter_elevation
//                } else {
//                    if (isPlatCood) R.string.please_input_z else R.string.please_enter_elevation
//                }
                ToastUtils.showShort(hint)
                return
            }
        }

        if (isNewCreatePoint()) {
            checkSameNamePoint(pointNum) {
                success()
            }
            return
        }
        success()
    }

    private fun checkSameNamePoint(pointNum: String, success: () -> Unit) {
        lifecycleScope.launch {
            val pointRecord = withContext(Dispatchers.IO) {
                val pointDataSource = SurveyorDatabaseFactory.instance.getPointDataSource()
                pointDataSource.findByNameSync(pointNum, getProjectId())
            }

            if (pointRecord != null) {
                //点名重复，提示
                TipsFragment.tipsOnly(
                    supportFragmentManager,
                    this@CreateEditPointActvity,
                    this@CreateEditPointActvity,
                    "同一项目下点号不能重复，请修改后重试操作。"
                )
                return@launch
            }
            success()
        }
    }

    /**
     * 设置平面还是大地坐标
     */
    private fun setBaseUi() {
        if (!isNewCreatePoint()) {
            mBinding.tvSaveContinue.visibility = View.GONE
        }

        val projectType = getProjectType()
        if (isLevelProject()) {
            mBinding.ilPointX.root.visibility = View.GONE
            mBinding.ilPointY.root.visibility = View.GONE

            setBaseTitleHint(
                R.string.x_coord,
                R.string.y_coord,
                R.string.elevation,
                R.string.please_input_x,
                R.string.please_input_y,
                R.string.please_enter_elevation
            )
        } else if (ProjectType.isHeightTraverse(projectType)) {
            setBaseTitleHint(
                R.string.x_coord,
                R.string.y_coord,
                R.string.elevation,
                R.string.please_input_x,
                R.string.please_input_y,
                R.string.please_enter_elevation
            )
        } else {
            setBaseTitleHint(
                R.string.x_coord, R.string.y_coord, R.string.z_coord,
                R.string.please_input_x, R.string.please_input_y, R.string.please_input_z
            )
        }

        setEditData()
    }

    private fun setBaseTitleHint(
        @StringRes xCoord: Int, yCoord: Int, zCoord: Int,
        pleaseInputX: Int, pleaseInputY: Int, @StringRes pleaseInputZ: Int
    ) {
        mBinding.ilPointX.tvTitle.setText(xCoord)
        mBinding.ilPointY.tvTitle.setText(yCoord)
        mBinding.ilPointZ.tvTitle.setText(zCoord)
        mBinding.ilPointX.editText.setHint(pleaseInputX)
        mBinding.ilPointY.editText.setHint(pleaseInputY)
        mBinding.ilPointZ.editText.setHint(pleaseInputZ)
    }

    private fun setEditData() {
        editPoint?.let {
            mBinding.ilPointNum.editText.setText(it.getName())

            setCoordData(it.getX(), it.getY(), it.getZ())
//            setCoordData(it.getLongitude(), it.getLatitude(), it.getZ())
        }
    }

    private fun setCoordData(x: Double, y: Double, z: Double) {
        mBinding.ilPointX.editText.setText(x.toString())
        mBinding.ilPointY.editText.setText(y.toString())
        mBinding.ilPointZ.editText.setText(z.toString())
    }

    override fun initData() {


    }

    override fun initToolBar() {
        initIntentData()
        val title = if (editPoint == null) R.string.new_point else R.string.edit_point
        setToolbarTitle(title)
        super.initToolBar()
    }

    private fun getProjectId() = intent.getStringExtra(BundleConstants.KEY_PROJECT_ID)!!
    private fun getProjectType() = intent.getStringExtra(BundleConstants.KEY_PROJECT_TYPE)!!

    private fun getPointType() = intent.getStringExtra(KEY_POINT_TYPE)!!

    private fun getEditPoint() =
        intent.getParcelableExtra<PointRecord>(KEY_EDIT_POINT)

    private fun isNewCreatePoint() = editPoint == null

    private fun getPointNumber() = mBinding.ilPointNum.editText.text.toString().trim()
    private fun getX() = mBinding.ilPointX.editText.text.toString().trim()
    private fun getY() = mBinding.ilPointY.editText.text.toString().trim()
    private fun getZ() = mBinding.ilPointZ.editText.text.toString().trim()

    private fun getXDouble(): Double {
        return xToDouble(getX())
    }

    private fun getYDouble(): Double {
        return xToDouble(getY())
    }

    private fun getZDouble(): Double {
        return xToDouble(getZ())
    }

    private fun xToDouble(value: String): Double {
        return if (TextUtils.isEmpty(value)) {
            0.0
        } else {
            value.toDouble()
        }
    }

    private fun clearContent() {
        mBinding.ilPointX.editText.setText("")
        mBinding.ilPointY.editText.setText("")
        mBinding.ilPointZ.editText.setText("")
        mBinding.ilPointX.editText.requestFocus()
    }

    private fun getNextPointNum(): String {
        val pointNumber = getPointNumber()
        val index = getIndexForTianjin(pointNumber)
        val word = getWordForTianjin(pointNumber)
        return "${word}${index + 1}"
    }

    private fun getIndexForTianjin(pointName: String): Int {
        val pre = "\\D*"
        val data = pointName.replace(pre.toRegex(), "")
        return try {
            data.toInt()
        } catch (e: NumberFormatException) {
            0
        }
    }

    private fun getWordForTianjin(pointName: String): String {
        val pre = "\\d*"
        return pointName.replace(pre.toRegex(), "")
    }

    private fun saveRecordForEdit(pointRecord: PointRecord, success: () -> Unit) {
        lifecycleScope.launch(Dispatchers.IO) {
            val projectType = getProjectType()
            if (ProjectType.isLevel(projectType)) {
                pointRecord.setLongitude(getXDouble())
                pointRecord.setLatitude(getYDouble())
            } else {
                pointRecord.setX(getXDouble())
                pointRecord.setY(getYDouble())
            }
            pointRecord.setZ(getZDouble())
            pointRecord.setName(getPointNumber())
            pointRecord.setCode("")
            pointRecord.setRemarks("")
            SurveyorDatabaseFactory.instance
                .getPointDataSource()
                .saveSync(pointRecord)
            success()
        }
    }

    private fun saveRecordForCreate(success: (() -> Unit)? = null) {
        lifecycleScope.launch(Dispatchers.IO) {
            val pointRecord = createPointRecord(
                getProjectId(), getPointType(), getPointNumber(), "",
                getXDouble(), getYDouble(), getZDouble(), "", -1,
            )
            SurveyorDatabaseFactory.instance
                .getPointDataSource()
                .save(pointRecord)
            success?.invoke()
        }
    }

    @WorkerThread
    private fun createPointRecord(
        projectId: String,
        pointType: String,
        name: String,
        code: String,
        x: Double,
        y: Double,
        z: Double,
        remarks: String,
        orderIndex: Int,
    ): PointRecord {
        val projectType = getProjectType()
        return if (ProjectType.isLevel(projectType)) {
            PointRecord.createLlhPoint(projectId, pointType, name, x, y, z)
        } else {
            PointRecord.point(
                projectId, pointType, name, code,
                x, y, z, remarks,
                orderIndex.toLong()
            )
        }
    }

    private fun isLevelProject() = ProjectTypeValue.LEVEL == getProjectType()

}