package com.bingce.controlnetwork.newui.stationsetting

import androidx.lifecycle.ViewModel
import androidx.lifecycle.viewModelScope
import blankj.utilcode.util.ToastUtils
import com.bingce.controlapphelper.datasource.database.ProjectType
import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory
import com.bingce.controlapphelper.datasource.database.config.ConfigConstants
import com.bingce.controlapphelper.datasource.database.surveyorstation.SurveyorStationRecord
import com.bingce.controlapphelper.util.StationUtil
import com.bingce.controlnetwork.newui.stationsetting.model.StationSettingEnvironmentData
import com.bingce.controlnetwork.newui.stationsetting.model.StationSettingHeightData
import com.bingce.controlnetwork.newui.stationsetting.model.StationSettingSettingData
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext

class StationSettingVm : ViewModel() {

    val environmentData = StationSettingEnvironmentData()
    val heightData = StationSettingHeightData()
    val settingData = StationSettingSettingData()

    private lateinit var stationId: String
    var stationRecord: SurveyorStationRecord? = null

    //默认有测量数据
    var hasSurveyorData = true


    fun getLastEnvironmentData(callBack: () -> Unit) {
        viewModelScope.launch {
            withContext(Dispatchers.IO) {
                environmentData.getLastData()
            }
            if (isHasNotEnvironmentData()) {
                ToastUtils.showShort("没有上次数据")
                return@launch
            }
            callBack()
        }
    }

    /**
     * 获取默认的环境数据
     */
    fun getDefaultEnvironmentData() {
        environmentData.getStandardData()
    }

    fun getEnvironmentData(callBack: () -> Unit) {
        viewModelScope.launch {
            withContext(Dispatchers.IO) {
                environmentData.getLastData()
            }
            if (isHasNotEnvironmentData()) {
                getDefaultEnvironmentData()
            }
            callBack()
        }
    }

    //没有环境数据（只判断温度就够了）
    private fun isHasNotEnvironmentData() = environmentData.temperature == null


    fun initStationId(stationId: String, callBack: () -> Unit) {
        this.stationId = stationId
        viewModelScope.launch(Dispatchers.IO) {
            hasSurveyorData = StationUtil.hasSurveyorData(stationId)
            withContext(Dispatchers.Main) {
                callBack()
            }
        }
    }

    /**
     * 获取测站相关数据的入口
     */
    fun getStationRecord(callBack: () -> Unit) {
        if (stationRecord != null) {
            callBack()
            return
        }
        synchronized(this) {
            viewModelScope.launch(Dispatchers.IO) {
                if (stationRecord == null) {
                    stationRecord = SurveyorDatabaseFactory.instance.getSurveyorStation()
                        .getRecordSync(stationId)
                    //初始化高度信息
                    heightData.initData(stationRecord)
                    //初始化测站设置信息
                    settingData.initData(stationRecord)
                }

                if (stationRecord != null) {
                    withContext(Dispatchers.Main) {
                        callBack()
                    }
                }
            }
        }
    }

    fun setEnvironmentDataToStation(
        temperature: String,
        humidity: String,
        airPressure: String,
    ) {
        stationRecord?.let {
            it.temperature = temperature
            it.humidity = humidity
            it.airPressure = airPressure
        }
    }

    fun setPointDataToStation(pointIndex: Int, pointHeight: String) {
        stationRecord?.let {
            heightData.updatePointData(pointIndex, pointHeight)
        }
    }

    /**
     * 更新测点的状态
     * 当有一个点是单独测距这个测站就是单独测距
     */
    fun setPointSettingToStation(
        pointIndex: Int,
        aloneMeasureDistance: Boolean,
        selectionPrismType: Int,
        isOpenAtrNot: Boolean
    ) {
        stationRecord?.let {
            settingData.updatePointSettingData(
                pointIndex,
                aloneMeasureDistance,
                selectionPrismType,
                isOpenAtrNot
            )
        }
    }

    fun setStationHeightToStation(instrumentHeight: String) {
        stationRecord?.let {
            it.instrumentHeight = instrumentHeight
        }
    }

    fun setStationLeftRightToStation(leftRight: Boolean) {
        stationRecord?.let {
            it.isLeftAndRightSurveyor = leftRight
        }
    }

    fun getStationLeftRight() = stationRecord?.leftAndRightSurveyor ?: false
    fun getShowSlop() = stationRecord?.showSlopDistance ?: true
    fun setStationShowSlop(showSlopDistance: Boolean) {
        stationRecord?.let {
            it.showSlopDistance = showSlopDistance
        }
    }

    fun setStationZeroToStation(zero: Boolean) {
        stationRecord?.let {
            it.autoClose = zero
        }
    }

    fun getStationZero() = stationRecord?.autoClose ?: false

    fun getStationSurveyorPointSize() = stationRecord?.items?.size ?: 0

    /**
     * 存储数据库
     */
    fun saveToDataBase(callBack: () -> Unit) {
        if (stationRecord == null) return
        viewModelScope.launch {
            withContext(Dispatchers.IO) {
                saveLastConfigData()
                SurveyorDatabaseFactory.instance.surveyorStation.update(stationRecord)
            }
            callBack()
        }
    }

    private fun saveLastConfigData() {
        val configDataSource = SurveyorDatabaseFactory.instance.getConfigDataSource()
        //环境信息
        configDataSource.setSync(
            ConfigConstants.KEY_LAST_TEMPERATURE,
            stationRecord!!.getTemperature()
        )
        configDataSource.setSync(
            ConfigConstants.KEY_LAST_AIR_PRESSURE,
            stationRecord!!.getAirPressure()
        )
        configDataSource.setSync(
            ConfigConstants.KEY_LAST_HUMIDITY,
            stationRecord!!.getHumidity()
        )
    }

    fun isCanShowAloneDistance(): Boolean {
        if (stationRecord == null) {
            return false
        }
        return ProjectType.isConnection(stationRecord!!.projectType)
    }

//    /**
//     * 当测点的状态更新完再调用这个方法
//     */
//    fun updateSeparetionToStation() {
//        stationRecord?.let {
//            var aloneDistance = false
//            for (item in it.items) {
//                if (item.isAloneDistance) {
//                    aloneDistance = true
//                    break
//                }
//            }
//            it.aloneMeasureDistance = aloneDistance
//        }
//    }


}