package com.bingce.controlnetwork.surveyor.method;

import android.content.Context;
import android.os.Environment;
import android.text.TextUtils;

import androidx.annotation.WorkerThread;

import com.bingce.controlapphelper.datasource.database.ProjectType;
import com.bingce.controlapphelper.datasource.database.ProjectTypeValue;
import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory;
import com.bingce.controlapphelper.datasource.database.line.ILineRecordDataSource;
import com.bingce.controlapphelper.datasource.database.line.LineRecord;
import com.bingce.controlapphelper.datasource.database.point.IPointDataSource;
import com.bingce.controlapphelper.datasource.database.point.PointRecord;
import com.bingce.controlapphelper.datasource.database.project.IProjectDataSource;
import com.bingce.controlapphelper.datasource.database.project.ProjectRecord;
import com.bingce.controlapphelper.datasource.database.schedule.ISurveyorScheduleDataSource;
import com.bingce.controlapphelper.datasource.database.surveyorstation.ISurveyorStationDataSource;
import com.bingce.controlapphelper.datasource.database.surveyorstation.SurveyorStationRecord;
import com.bingce.controlapphelper.datasource.database.surveyorstation.model.SurveyorPoint;
import com.bingce.controlapphelper.datasource.database.tolerance.detail.IToleranceDetailDataSource;
import com.bingce.controlapphelper.datasource.database.tolerance.detail.ToleranceDetailRecord;
import com.bingce.controlapphelper.datasource.database.tolerance.group.ToleranceGroupRecord;
import com.bingce.controlapphelper.datasource.database.tolerance.level.ToleranceLevelRecord;
import com.bingce.controlapphelper.surveyor.data.CachedData;
import com.bingce.controlapphelper.surveyor.data.CachedDataAloneDistance;
import com.bingce.controlapphelper.surveyor.method.ExportStationMethod;
import com.bingce.controlapphelper.surveyor.method.ExportUtils;
import com.bingce.controlapphelper.surveyor.method.SurveyorAdjustMethod;
import com.bingce.controlapphelper.util.Tools;
import com.bingce.controlnetwork.App;
import com.bingce.controlnetwork.R;
import com.bingce.controlnetwork.newui.configfragment.SelectExportType;
import com.bingce.controlnetwork.util.CheckRegUtil;
import com.bingce.utils.DateUtils;
import com.bingce.utils.RegisterUtil;
import com.bingce.utils.Util;
import com.bingce.utils.VipManager;
import com.bingce.zip.ZipUtil;

import org.jetbrains.annotations.NotNull;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import blankj.utilcode.util.FileUtils;
import blankj.utilcode.util.ToastUtils;


public class ExportUtilsPla {
    private final static String NEW_LINE = "\r\n";
    private final static String SPACE = "\t";

    private final static int ANGLE_DECIMALS = 5; //保留小数位数
    public final static int DISTANCE_DECIMALS = 5; //距离保留小数位数
    public final static int PPM_DECIMALS = 8; //ppm保留小数位数

    public static String getExportFirstDirectory() {
        return App.getApp().getString(R.string.app_name) + "导出文件";
    }

    public static String getExportSecondDirectory() {
        return getExportFirstDirectory() + File.separator + "导线文件导出";
    }

    /**
     * 收集需要导出的测站、测期、项目相关信息
     */
    @WorkerThread
    private static Collection<Project> collect(List<ExportItem> exportItems) {
        Map<String, Project> projectMap = new HashMap<>();
        for (ExportItem exportItem : exportItems) {
            Project project;
            if (!projectMap.containsKey(exportItem.projectId)) {
                project = new Project(exportItem.projectId, exportItem.projectType, exportItem.projectName);
                projectMap.put(exportItem.projectId, project);
            } else {
                project = projectMap.get(exportItem.projectId);
            }
            if (project == null) {
                continue;
            }
            SurveyorPeriod cycle;
            if (!project.surveyorPeriodMap.containsKey(exportItem.cycleId)) {
                cycle = new SurveyorPeriod(exportItem.cycleId, exportItem.cycleName);
                project.surveyorPeriodMap.put(exportItem.cycleId, cycle);
            } else {
                cycle = project.surveyorPeriodMap.get(exportItem.cycleId);
            }

            if (cycle == null) {
                continue;
            }
            if (!cycle.stationMap.containsKey(exportItem.stationId)) {
                Station station = new Station(exportItem.stationId, exportItem.stationPointName);
                cycle.stationMap.put(exportItem.stationId, station);
            }
        }
        return projectMap.values();
    }

    protected static List<PointRecord> surveyorPointRecordListInAllPeriods(Project project) {
        //收集所有测期下所有测站的测量点记录，用于后面导出控制点
        List<PointRecord> surveyorPointRecordListInAllPeriods = new ArrayList<>();
        for (Map.Entry<String, SurveyorPeriod> surveyorPeriodEntry : project.surveyorPeriodMap.entrySet()) {
            //遍历每个测期
            SurveyorPeriod surveyorPeriod = surveyorPeriodEntry.getValue();
            for (Map.Entry<String, Station> stationEntry : surveyorPeriod.stationMap.entrySet()) {
                ISurveyorStationDataSource stationDataSource = SurveyorDatabaseFactory.instance.getSurveyorStation();
                SurveyorStationRecord stationRecord = stationDataSource.getRecordSync(stationEntry.getValue().stationId);
                if (stationRecord == null) {
                    continue;
                }
                //收集测站中的测点记录
                List<PointRecord> surveyorPointListInSingleStation = ExportStationMethod.surveyorPointRecordInSingleStation(stationRecord);
                for (PointRecord record : surveyorPointListInSingleStation) {
                    //去重.检测对应测点是否已经记录到列表中
                    boolean hasRecord = false;
                    for (PointRecord savedRecord : surveyorPointRecordListInAllPeriods) {
                        if (record.id.equals(savedRecord.id)) {
                            hasRecord = true;
                            break;
                        }
                    }
                    if (hasRecord) {
                        continue;
                    }
                    surveyorPointRecordListInAllPeriods.add(record);
                }
            }
        }
        return surveyorPointRecordListInAllPeriods;
    }

    @WorkerThread
    private static List<PointRecord> collectPointFromProject(Project project) {
        IPointDataSource pointDataSource = SurveyorDatabaseFactory.instance.getPointDataSource();
        List<PointRecord> pointList = pointDataSource.findByProjectIdSync(project.projectId);
        //按照已知点在前未知点在后排序
        Collections.sort(pointList, (current, Other) -> {
            if ((PointRecord.TYPE_KNOWN_POINT.equals(current.pointType) || PointRecord.TYPE_GLOBAL_POINT.equals(current.pointType))
                    && PointRecord.TYPE_UNKNOWN_POINT.equals(Other.pointType)) {
                //当前是已知点 比较是未知点
                return 1;
            } else if (PointRecord.TYPE_UNKNOWN_POINT.equals(current.pointType)
                    && (PointRecord.TYPE_KNOWN_POINT.equals(Other.pointType) || PointRecord.TYPE_GLOBAL_POINT.equals(Other.pointType))) {
                //当前是未知点，比较是已知点
                return -1;
            }
            return 0;
        });
        return pointList;
    }

    /**
     * 导出冰测数据
     */
    @WorkerThread
    private static void exportBingCe(Project project, ToleranceDetailRecord toleranceRecord, File projectCachedDir) {

//        String controlPointTxtPrint = getControlPointTxtPrint(project);
//        String projectTxtPrint = getProjectTxtPrint(project);

        //读取「限差」信息
        //二级
        ToleranceLevelRecord toleranceLevelRecord = null;
        if (toleranceRecord.levelId != null) {
            toleranceLevelRecord = SurveyorDatabaseFactory.instance.getToleranceLevelDataSource().getRecordSync(toleranceRecord.levelId);
        }
        //一级
        ToleranceGroupRecord toleranceGroupRecord = null;
        if (toleranceRecord.groupId != null) {
            toleranceGroupRecord = SurveyorDatabaseFactory.instance.getToleranceGroupDataSource().getRecordSync(toleranceRecord.groupId);
        }

        //遍历每个测期
        for (Map.Entry<String, SurveyorPeriod> surveyorPeriodEntry : project.surveyorPeriodMap.entrySet()) {
            //初始化天津水平角，垂直角和距离
            StringBuilder horSbTianjinOne = new StringBuilder();
            StringBuilder verSbTianjinOne = new StringBuilder();
            StringBuilder disSbTianjinOne = new StringBuilder();

            SurveyorPeriod surveyorPeriod = surveyorPeriodEntry.getValue();
            //测站文件夹和测站文件的名称暂存map
            Map<String, Integer> stationFileNameMap = new HashMap<>();
            //创建测期文件夹
            File periodFolder = ExportStationMethod.initPeriodFolder(projectCachedDir, surveyorPeriod.periodName);

            for (Map.Entry<String, Station> stationEntry : surveyorPeriod.stationMap.entrySet()) {
                String stationName = stationEntry.getValue().stationName;
                String stationFileName = getStationFileName(stationFileNameMap, stationName);

                ISurveyorStationDataSource stationDataSource = SurveyorDatabaseFactory.instance.getSurveyorStation();
                SurveyorStationRecord stationRecord = stationDataSource.getRecordSync(stationEntry.getValue().stationId);
                if (stationRecord == null) {
                    continue;
                }
                //创建测站文件夹
                File stationFolder = ExportStationMethod.initStationFolder(periodFolder, stationFileName);

                //2.导出单个测站的数据
                ExportStationMethod.exportBingCeSingleStationSurveyorData(
                        stationFolder,
                        toleranceRecord, toleranceLevelRecord, toleranceGroupRecord,
                        stationFileName, stationName, stationRecord);
                //天津收集方向值，垂直角，边长文件的数据
                collectCustomTj61365(stationName, stationRecord, toleranceRecord, horSbTianjinOne, verSbTianjinOne, disSbTianjinOne);
            }

            //输出文件
            if (horSbTianjinOne.length() > 0) {
                horSbTianjinOne.append("END");
                ExportStationMethod.save(horSbTianjinOne.toString(), periodFolder, "方向值.TXT");
            } else {
                ExportStationMethod.save("   ", periodFolder, "方向值.TXT");
            }
            if (verSbTianjinOne.length() > 0) {
                ExportStationMethod.save(verSbTianjinOne.toString(), periodFolder, "垂直角.TXT");
            } else {
                ExportStationMethod.save("   ", periodFolder, "垂直角.TXT");
            }
            if (disSbTianjinOne.length() > 0) {
                disSbTianjinOne.append("END");
                ExportStationMethod.save(disSbTianjinOne.toString(), periodFolder, "边长.TXT");
            } else {
                ExportStationMethod.save("   ", periodFolder, "边长.TXT");
            }

            exportProjectAndControlPoint(project,projectCachedDir);
        }

    }

    /**
     * 导出项目和控制点文件
     *
     * @param project
     * @param projectDir
     */
    private static void exportProjectAndControlPoint(Project project, File projectDir) {
        //存储已知点
        List<PointRecord> surveyorPointRecordListInAllPeriods = surveyorPointRecordListInAllPeriods(project);
        StringBuilder pointStringBuilder = new StringBuilder();
        for (PointRecord pointRecord : surveyorPointRecordListInAllPeriods) {
            if (pointRecord == null) {
                continue;
            }
            if (PointRecord.TYPE_KNOWN_POINT.equals(pointRecord.pointType) ||
                    PointRecord.TYPE_GLOBAL_POINT.equals(pointRecord.pointType)) {
                pointStringBuilder
                        .append(pointRecord.getName())
                        .append(",")
                        .append(pointRecord.getXFormat(4))
                        .append(",")
                        .append(pointRecord.getYFormat(4))
                        .append(",")
                        .append(pointRecord.getZFormat(4))
                        .append("\n");
            }
        }
        //导出控制点文件
        ExportStationMethod.save(pointStringBuilder.toString(), projectDir, "controlpoint.TXT");
        //导出项目名称和项目类型文件
        ExportStationMethod.save(project.projectName + "," + project.projectType + "\n", projectDir, "project.TXT");
    }

    private static String getControlPointTxtPrint(Project project) {
        List<PointRecord> surveyorPointRecordListInAllPeriods = collectPointFromProject(project);
        StringBuilder pointStringBuilder = new StringBuilder();

        for (PointRecord pointRecord : surveyorPointRecordListInAllPeriods) {
            if (pointRecord == null) {
                continue;
            }

            pointStringBuilder
                    .append(pointRecord.getName())
                    .append(",")
                    .append("")
                    .append(",")
                    .append(ExportStationMethod.getXyzFormat(pointRecord.getX()))
                    .append(",")
                    .append(ExportStationMethod.getXyzFormat(pointRecord.getY()))
                    .append(",")
                    .append(ExportStationMethod.getXyzFormat(pointRecord.getZ()))
                    .append(",")
                    .append(pointRecord.getLongitude())
                    .append(",")
                    .append(pointRecord.getLatitude())
                    .append(",")
                    .append(getPointTypeForTianjin(pointRecord.getPointType()));

            pointStringBuilder.append("\n");
        }
        return pointStringBuilder.toString();
    }

    @WorkerThread
    private static String getProjectTxtPrint(Project project) {
        IProjectDataSource projectDataSource = SurveyorDatabaseFactory.instance.getProjectDataSource();
//        ProjectRecord projectRecord = projectDataSource.getRecord(project.projectId);

        return project.projectName +
                "," +
                getProjectTypeForTianjin(project.projectType) +
                "\n";

    }


    /**
     * nameMap主要是为了避免同一测期下测站重复的问题
     * 比如闭合导线，存在某点充当多次站点的情况
     * 通过nameMap为后面的station添加一个'-x'后缀
     *
     * @param nameMap
     * @param stationName
     * @return
     */
    private static String getStationFileName(Map<String, Integer> nameMap, String stationName) {
        if (nameMap.containsKey(stationName)) {
            Integer obj = nameMap.get(stationName);
            if (obj != null) {
                int count = obj + 1;
                nameMap.put(stationName, count);
                stationName = stationName + "_" + count;
            }
        } else {
            nameMap.put(stationName, 0);
        }
        return stationName;
    }

    /**
     * 收集边长，垂直角，方向值文件的数据
     *
     * @param stationName
     * @param stationRecord
     * @param toleranceRecord
     * @param horSbTianjinOne
     * @param verSbTianjinOne
     * @param disSbTianjinOne
     */
    @WorkerThread
    private static void collectCustomTj61365(String stationName, SurveyorStationRecord stationRecord, ToleranceDetailRecord toleranceRecord,
                                             StringBuilder horSbTianjinOne, StringBuilder verSbTianjinOne,
                                             StringBuilder disSbTianjinOne) {
//        if (stationRecord.isSeparateObservation()) {
//            collectAloneHorAngleTjUnique(stationName, stationRecord, toleranceRecord, horSbTianjinOne);
//        } else {
//            collectCombineAngleDistanceTjUnique(stationName, stationRecord, toleranceRecord, horSbTianjinOne, verSbTianjinOne, disSbTianjinOne);
//        }
        collectCombineAngleDistanceTjUnique(stationName, stationRecord, toleranceRecord, horSbTianjinOne, verSbTianjinOne, disSbTianjinOne);
        collectAloneDistanceTjUnique(stationName, stationRecord, toleranceRecord, disSbTianjinOne);
    }

    /**
     * 收集合并观测的水平角，垂直角和距离
     * 距离可能存在单独测距
     */
    private static void collectCombineAngleDistanceTjUnique(String stationName, SurveyorStationRecord stationRecord, ToleranceDetailRecord toleranceRecord,
                                                            StringBuilder horSbTianjinOne, StringBuilder verSbTianjinOne,
                                                            StringBuilder disSbTianjinOne) {
        CachedData cachedData =
                CachedData.newInstance(stationRecord.isAutoClose(),
                        stationRecord.isLeftAndRightSurveyor(),
                        stationRecord.getId(), toleranceRecord);

        //测距测角的平差结果
        SurveyorAdjustMethod.AdjustResult adjustResult = SurveyorAdjustMethod.calStationAdjust(stationRecord, cachedData);
        if (adjustResult.errorString == null || adjustResult.errorString.isEmpty()) {
            //水平角
            horSbTianjinOne.append(stationName).append(SPACE).append(formatHeight(Double.parseDouble(stationRecord.getInstrumentHeight()))).append(SPACE);

            StringBuilder verSbTianjinEveryStation = new StringBuilder();
            int verEverySurveyorPoint = 0;
            //遍历测站测点
            for (int i = 0; i < adjustResult.result.size(); i++) {
                SurveyorAdjustMethod.Item item = adjustResult.result.get(i);
                //水平角
                String exportHorAngle = ExportStationMethod.radianToDmsForTj61365(item.getHorAngle(), true);
                if (i != 0) {
                    horSbTianjinOne.append(SPACE).append(SPACE);
                }
                horSbTianjinOne.append(item.getPointName()).append(SPACE).append(exportHorAngle).append(SPACE).append(formatHeight(item.getPrismHeight()));
                horSbTianjinOne.append(NEW_LINE);

                verEverySurveyorPoint++;
                String exportVerAngle = ExportStationMethod.radianToDmsForTj61365(item.getVerAngle(), false);
                verSbTianjinEveryStation.append(item.getPointName()).append(SPACE).append(exportVerAngle).append(SPACE).append(formatHeight(Double.parseDouble(stationRecord.getInstrumentHeight())))
                        .append(SPACE).append(formatHeight(item.getPrismHeight()));
                verSbTianjinEveryStation.append(NEW_LINE);

                double distance = item.getDistance();
                if (distance > 0) {
                    appendAloneDistance(disSbTianjinOne, stationName, item.getPointName(), distance,
                            Double.parseDouble(stationRecord.getInstrumentHeightDeal()), String.valueOf(item.getPrismHeight()));
                }
            }

            if (verEverySurveyorPoint > 0) {
                verSbTianjinOne.append(stationName).append(SPACE).append(verEverySurveyorPoint).append(NEW_LINE);
                verSbTianjinOne.append(verSbTianjinEveryStation);
            }

        }
    }

    /**
     * 收集天津边长数据
     * 合并观测和分开观测的单独测距
     */
    private static void collectAloneDistanceTjUnique(String stationName, SurveyorStationRecord stationRecord,
                                                     ToleranceDetailRecord toleranceRecord, StringBuilder disSbTianjinOne) {
        SurveyorAdjustMethod.calAloneDistance(stationRecord.getId(),
                CachedDataAloneDistance.newInstance(stationRecord.getId()), new SurveyorAdjustMethod.ICallback() {
                    @Override
                    public void onGet(ArrayList<SurveyorAdjustMethod.Item> adjustResult, Boolean isShowSlopDistance) {
                        if (!adjustResult.isEmpty()) {
                            for (SurveyorAdjustMethod.Item item : adjustResult) {
                                appendAloneDistance(disSbTianjinOne, stationName, item.getPointName(), item.getDistance(),
                                        Double.parseDouble(stationRecord.getInstrumentHeightDeal()), String.valueOf(item.getPrismHeight()));
                            }
                        }
                    }

                    @Override
                    public void onFailure(String error) {

                    }
                });
    }

    private static void appendAloneDistance(StringBuilder disSbTianjinOne, String stationName, String pointName, double distance,
                                            double instrumentHeight, String prismHeight) {
        disSbTianjinOne.append(stationName).append(SPACE).append(pointName).append(SPACE)
                .append(Tools.formatDouble2String(distance, 4))
                .append(SPACE).append(formatHeight(instrumentHeight))
                .append(SPACE).append(formatHeight(Double.parseDouble(prismHeight)));
        disSbTianjinOne.append(NEW_LINE);
    }

    /**
     * 导出科傻数据
     */
    @WorkerThread
    private static void exportKeSha(Project project,
                                    ToleranceDetailRecord toleranceRecord, File projectCachedDir) {
        ISurveyorStationDataSource stationDataSource = SurveyorDatabaseFactory.instance.getSurveyorStation();
        StringBuilder ksStringBuilder = new StringBuilder();
        ksStringBuilder.append("3.5,5,5").append(NEW_LINE);//限差使用默认值

        //存储已知点
        List<PointRecord> surveyorPointRecordListInAllPeriods = surveyorPointRecordListInAllPeriods(project);
        for (PointRecord pointRecord : surveyorPointRecordListInAllPeriods) {
            if (pointRecord == null) {
                continue;
            }
            if (PointRecord.TYPE_KNOWN_POINT.equals(pointRecord.pointType) ||
                    PointRecord.TYPE_GLOBAL_POINT.equals(pointRecord.pointType)) {
                ksStringBuilder
                        .append(pointRecord.getName())
                        .append(",")
                        .append(pointRecord.getX())
                        .append(",")
                        .append(pointRecord.getY())
                        .append(NEW_LINE);
            }
        }
        //存储其他数据
        for (Map.Entry<String, SurveyorPeriod> cycleEntry : project.surveyorPeriodMap.entrySet()) {
            SurveyorPeriod cycle = cycleEntry.getValue();
            String cycleName = cycle.periodName.replace(":", "_");
            File cycleDir = new File(projectCachedDir, cycleName);
            try {
                if (!cycleDir.exists()) {
                    cycleDir.mkdir();
                }
            } catch (Exception ignore) {
                continue;
            }
            for (Map.Entry<String, Station> stationEntry : cycle.stationMap.entrySet()) {
                String stationId = stationEntry.getValue().stationId;
                String stationName = stationEntry.getValue().stationName;
                //读取测站信息
                SurveyorStationRecord stationRecord = stationDataSource.getRecordSync(stationId);

                //创建缓存数据
                CachedData cachedData =
                        CachedData.newInstance(stationRecord.isAutoClose(),
                                stationRecord.isLeftAndRightSurveyor(),
                                stationId, toleranceRecord);
                if (cachedData == null) {
                    continue;
                }
                cachedData.onShowSlopDistance(false);//平距

                //计算平差数据
                SurveyorAdjustMethod.AdjustResult adjustResult = SurveyorAdjustMethod.calStationAdjust(stationRecord, cachedData);
                if (TextUtils.isEmpty(adjustResult.errorString)) {
                    ksStringBuilder.append(stationName).append(NEW_LINE);
                    for (SurveyorAdjustMethod.Item item : adjustResult.result) {
                        ksStringBuilder
                                .append(item.getPointName())
                                .append(",L,")
                                .append(ExportStationMethod.radianToDmsDouble(item.getHorAngle()))
                                .append(NEW_LINE);
                    }

                    for (SurveyorAdjustMethod.Item item : adjustResult.result) {
                        //这里兼容下钢丝测距的距离
                        String pointName = item.getPointName();

                        double distance = 0;

                        if (!TextUtils.equals(ProjectTypeValue.CONNECTION, stationRecord.projectType)) {
                            distance = item.getDistance();
                        } else {
                            Map<String, Double> wellAverageDistanceMap = ExportStationMethod.getWellAverageDistanceFromStation(stationRecord);
                            if (wellAverageDistanceMap != null && wellAverageDistanceMap.containsKey(pointName)) {
                                distance = wellAverageDistanceMap.get(pointName);
                            }
                            if (distance == 0) {
                                distance = item.getDistance();
                            }
                        }

                        ksStringBuilder
                                .append(pointName)
                                .append(",S,")
                                .append(distance)
                                .append(NEW_LINE);
                    }
                }


            }
            ExportStationMethod.save(ksStringBuilder.toString(), cycleDir, cycleName + ".in2");
        }
    }

    /**
     * 导出南方平差仪数据 .txt文件
     */
    @WorkerThread
    private static void exportSouthAdjustment(Project project,
                                              ToleranceDetailRecord toleranceRecord, File projectCachedDir) {
        ISurveyorStationDataSource stationDataSource = SurveyorDatabaseFactory.instance.getSurveyorStation();
        StringBuilder printSb = new StringBuilder();

        //[NET]部分
        printSb.append(NEW_LINE);
        printSb.append("[NET]");
        printSb.append("Name:" + project.projectName);
        printSb.append(NEW_LINE);
        printSb.append("Organ:SOUTH SURVEY");
        printSb.append(NEW_LINE);
        printSb.append("Obser:SOUTH");
        printSb.append(NEW_LINE);
        printSb.append("Computer:SOUTH");
        printSb.append(NEW_LINE);
        printSb.append("Recorder:SOUTH");
        printSb.append(NEW_LINE);
        printSb.append("Remark:" + project.projectName);
        printSb.append(NEW_LINE);
        printSb.append("Software:南方平差易2002");
        printSb.append(NEW_LINE);
        //[PARA]部分
        printSb.append("[PARA]");
        printSb.append(NEW_LINE);
        printSb.append("MO:2.5");
        printSb.append(NEW_LINE);
        printSb.append("MS:5.0");
        printSb.append(NEW_LINE);
        printSb.append("MR:5.0");
        printSb.append(NEW_LINE);
        printSb.append("DistanceError:0.0000");
        printSb.append(NEW_LINE);
        printSb.append("DistanceMethod:0");
        printSb.append(NEW_LINE);
        printSb.append("LevelMethod:0");
        printSb.append(NEW_LINE);
        printSb.append("Mothed:1");
        printSb.append(NEW_LINE);
        printSb.append("LevelTrigon:0");
        printSb.append(NEW_LINE);
        printSb.append("TrigonObser:1");
        printSb.append(NEW_LINE);
        printSb.append("Times:1");
        printSb.append(NEW_LINE);
        printSb.append("Level:4");
        printSb.append(NEW_LINE);
        printSb.append("Level1:2");
        printSb.append(NEW_LINE);
        printSb.append("Limit:3");
        printSb.append(NEW_LINE);
        printSb.append("Format:1");
        printSb.append(NEW_LINE);
        //[STATION]部分
        printSb.append("[STATION]");
        printSb.append(NEW_LINE);
        //存储已知点
        List<PointRecord> surveyorPointRecordListInAllPeriods = surveyorPointRecordListInAllPeriods(project);
        for (PointRecord pointRecord : surveyorPointRecordListInAllPeriods) {
            if (pointRecord == null) {
                continue;
            }

            printSb.append(pointRecord.getName() + ",");

            switch (pointRecord.pointType) {
                case PointRecord.TYPE_KNOWN_POINT:
                    printSb.append("10");
                    break;
                default:
                    printSb.append("00");
                    break;
            }
            printSb.append(",");
            printSb.append(ExportStationMethod.getXyzFormat(pointRecord.getX()));
            printSb.append(",");
            printSb.append(ExportStationMethod.getXyzFormat(pointRecord.getY()));
            printSb.append(",,,,");
            printSb.append(NEW_LINE);
        }
        //[OBSER]部分
        printSb.append("[OBSER]");
        printSb.append(NEW_LINE);

        for (Map.Entry<String, SurveyorPeriod> cycleEntry : project.surveyorPeriodMap.entrySet()) {
            SurveyorPeriod cycle = cycleEntry.getValue();
            String cycleName = cycle.periodName.replace(":", "_");
            File cycleDir = new File(projectCachedDir, cycleName);
            try {
                if (!cycleDir.exists()) {
                    cycleDir.mkdir();
                }
            } catch (Exception ignore) {
                continue;
            }

            StringBuilder stationSb = new StringBuilder();

            for (Map.Entry<String, Station> stationEntry : cycle.stationMap.entrySet()) {
                String stationId = stationEntry.getValue().stationId;
                String stationName = stationEntry.getValue().stationName;
                //读取测站信息
                SurveyorStationRecord stationRecord = stationDataSource.getRecordSync(stationId);

                //创建缓存数据
                CachedData cachedData =
                        CachedData.newInstance(stationRecord.isAutoClose(),
                                stationRecord.isLeftAndRightSurveyor(),
                                stationId, toleranceRecord);
                if (cachedData == null) {
                    continue;
                }
                cachedData.onShowSlopDistance(false);

                //计算平差数据
                SurveyorAdjustMethod.AdjustResult adjustResult = SurveyorAdjustMethod.calStationAdjust(stationRecord, cachedData);
                if (TextUtils.isEmpty(adjustResult.errorString)) {
                    for (SurveyorAdjustMethod.Item item : adjustResult.result) {
                        //这里兼容下钢丝测距的距离
                        String pointName = item.getPointName();

                        double distance = 0;

                        if (!TextUtils.equals(ProjectTypeValue.CONNECTION, stationRecord.projectType)) {
                            distance = item.getDistance();
                        } else {
                            Map<String, Double> wellAverageDistanceMap = ExportStationMethod.getWellAverageDistanceFromStation(stationRecord);
                            if (wellAverageDistanceMap != null && wellAverageDistanceMap.containsKey(pointName)) {
                                distance = wellAverageDistanceMap.get(pointName);
                            }
                            if (distance == 0) {
                                distance = item.getDistance();
                            }
                        }

                        stationSb.append(stationName).append(",")
                                .append(item.getPointName()).append(",")
                                .append(ExportStationMethod.radianToDmsDouble(item.getHorAngle()))
                                .append(",")
                                .append(distance)
                                .append(",,,,,,,")
                                .append(NEW_LINE);
                    }
                }

            }

            printSb.append(stationSb);

            ExportStationMethod.save(printSb.toString(), cycleDir, cycleName + ".txt");
        }
    }

    /**
     * 导出启飞数据 .kzw文件
     */
    @WorkerThread
    private static void exportOpenFly(Project project,
                                      ToleranceDetailRecord toleranceRecord, File projectCachedDir) {
        ISurveyorStationDataSource stationDataSource = SurveyorDatabaseFactory.instance.getSurveyorStation();
        StringBuilder printSb = new StringBuilder();

        printSb.append("控制测量优化设计与平差计算文件");
        printSb.append(NEW_LINE);
        printSb.append("300");
        printSb.append(NEW_LINE);
        printSb.append("[全网属性]");
        printSb.append(NEW_LINE);
        printSb.append("网名:" + project.projectName);
        printSb.append(NEW_LINE);
        printSb.append("是否优化:0");
        printSb.append(NEW_LINE);
        printSb.append("HELMERT定权:0");
        printSb.append(NEW_LINE);
        printSb.append("测角中误差:2.500000");
        printSb.append(NEW_LINE);
        printSb.append("测距定权方式:0");
        printSb.append(NEW_LINE);
        printSb.append("测距A值:50.000000");
        printSb.append(NEW_LINE);
        printSb.append("测距B值:0.000000");
        printSb.append(NEW_LINE);
        printSb.append("高程类型:0");
        printSb.append(NEW_LINE);
        printSb.append("水准定权方式:0");
        printSb.append(NEW_LINE);
        printSb.append("三角高程测距形式:0");
        printSb.append(NEW_LINE);
        printSb.append("是否由斜距计算平距:1");
        printSb.append(NEW_LINE);
        printSb.append("大气折光:0.140000");
        printSb.append(NEW_LINE);
        printSb.append("高程单位权:10.000000");
        printSb.append(NEW_LINE);
        printSb.append("坐标系:0");
        printSb.append(NEW_LINE);
        printSb.append("是否含500:0");
        printSb.append(NEW_LINE);
        printSb.append("是否高程归化:0");
        printSb.append(NEW_LINE);
        printSb.append("归算高程面:0.000000");
        printSb.append(NEW_LINE);
        printSb.append("图数更新方式:0");
        printSb.append(NEW_LINE);
        printSb.append("平面网等级:2");
        printSb.append(NEW_LINE);
        printSb.append("高程网等级:2");
        printSb.append(NEW_LINE);
        printSb.append("三角形限差:9.000000");
        printSb.append(NEW_LINE);
        printSb.append("方位角限差:5.000000");
        printSb.append(NEW_LINE);
        printSb.append("导线段相对闭合差限差:35000");
        printSb.append(NEW_LINE);
        printSb.append("高差闭合差限差:20.000000");
        printSb.append(NEW_LINE);
        printSb.append("角度取位:1");
        printSb.append(NEW_LINE);
        printSb.append("边长取位:3");
        printSb.append(NEW_LINE);
        printSb.append("高差取位:3");
        printSb.append(NEW_LINE);
        printSb.append("坐标取位:3");
        printSb.append(NEW_LINE);
        printSb.append("高程取位:3");
        printSb.append(NEW_LINE);
        printSb.append("已知点类型:0");
        printSb.append(NEW_LINE);
        printSb.append("未知点类型:3");
        printSb.append(NEW_LINE);
        printSb.append("图形比例:1000.000000");
        printSb.append(NEW_LINE);
        printSb.append("椭圆比例:5.000000");
        printSb.append(NEW_LINE);
        printSb.append("END");
        printSb.append(NEW_LINE);
        //获取点记录
        List<PointRecord> surveyorPointRecordListInAllPeriods = surveyorPointRecordListInAllPeriods(project);
        //已知点
        printSb.append("[已知点]");
        printSb.append(NEW_LINE);
        StringBuilder sbKnow = new StringBuilder();
        int knowCount = 0;
        for (PointRecord pointRecord : surveyorPointRecordListInAllPeriods) {
            if (pointRecord == null || !TextUtils.equals(pointRecord.pointType, PointRecord.TYPE_KNOWN_POINT)) {
                continue;
            }
            ++knowCount;
            sbKnow.append(pointRecord.getName() + ",");
            sbKnow.append(ExportStationMethod.getXyzFormat(pointRecord.getX()));
            sbKnow.append(",");
            sbKnow.append(ExportStationMethod.getXyzFormat(pointRecord.getY()));
            sbKnow.append(",,");
            sbKnow.append(NEW_LINE);
        }
        printSb.append(knowCount);
        printSb.append(NEW_LINE);
        printSb.append(sbKnow);
        printSb.append("END");
        printSb.append(NEW_LINE);
        //固定条件
        printSb.append("[固定条件]");
        printSb.append(NEW_LINE);
        printSb.append("0");
        printSb.append(NEW_LINE);
        printSb.append("END");
        printSb.append(NEW_LINE);
        //控制点
        printSb.append("[控制点]");
        printSb.append(NEW_LINE);
        StringBuilder sbControl = new StringBuilder();
        int controlCount = 0;
        for (PointRecord pointRecord : surveyorPointRecordListInAllPeriods) {
            if (pointRecord == null) {
                continue;
            }
            ++controlCount;
            sbControl.append(pointRecord.getName() + ",");
            sbControl.append(ExportStationMethod.getXyzFormat(pointRecord.getX()));
            sbControl.append(",");
            sbControl.append(ExportStationMethod.getXyzFormat(pointRecord.getY()));
            sbControl.append(",");
            sbControl.append(NEW_LINE);
        }
        printSb.append(controlCount);
        printSb.append(NEW_LINE);
        printSb.append(sbControl);
        printSb.append("END");
        printSb.append(NEW_LINE);
        //[测站数据]
        printSb.append("[测站数据]");
        printSb.append(NEW_LINE);

        for (Map.Entry<String, SurveyorPeriod> cycleEntry : project.surveyorPeriodMap.entrySet()) {
            SurveyorPeriod cycle = cycleEntry.getValue();
            String cycleName = cycle.periodName.replace(":", "_");
            File cycleDir = new File(projectCachedDir, cycleName);
            try {
                if (!cycleDir.exists()) {
                    cycleDir.mkdir();
                }
            } catch (Exception ignore) {
                continue;
            }

            StringBuilder stationSb = new StringBuilder();
            Map<String, List<SurveyorAdjustMethod.Item>> tempDataMap = new LinkedHashMap<>();

            for (Map.Entry<String, Station> stationEntry : cycle.stationMap.entrySet()) {
                String stationId = stationEntry.getValue().stationId;
                String stationName = stationEntry.getValue().stationName;
                //读取测站信息
                SurveyorStationRecord stationRecord = stationDataSource.getRecordSync(stationId);

                //创建缓存数据
                CachedData cachedData =
                        CachedData.newInstance(stationRecord.isAutoClose(),
                                stationRecord.isLeftAndRightSurveyor(),
                                stationId, toleranceRecord);
                if (cachedData == null) {
                    continue;
                }
                cachedData.onShowSlopDistance(false);

                //计算平差数据
                SurveyorAdjustMethod.AdjustResult adjustResult = SurveyorAdjustMethod.calStationAdjust(stationRecord, cachedData);
                if (TextUtils.isEmpty(adjustResult.errorString)) {
                    stationSb.append(cycle.stationMap.keySet().size());
                    stationSb.append(NEW_LINE);

                    stationSb.append(stationName).append(",").append(adjustResult.result.size()).append(",").append(NEW_LINE);
                    stationSb.append("照准点").append(NEW_LINE);

                    for (SurveyorAdjustMethod.Item item : adjustResult.result) {
                        //这里兼容下钢丝测距的距离
                        String pointName = item.getPointName();

                        double distance = 0;

                        if (!TextUtils.equals(ProjectTypeValue.CONNECTION, stationRecord.projectType)) {
                            distance = item.getDistance();
                        } else {
                            Map<String, Double> wellAverageDistanceMap = ExportStationMethod.getWellAverageDistanceFromStation(stationRecord);
                            if (wellAverageDistanceMap != null && wellAverageDistanceMap.containsKey(pointName)) {
                                distance = wellAverageDistanceMap.get(pointName);
                            }
                            if (distance == 0) {
                                distance = item.getDistance();
                            }
                        }

                        stationSb
                                .append(item.getPointName()).append(",")
                                .append(Tools.radianToDmsString(item.getHorAngle(), 2, true))
                                .append(",")
                                .append(distance)
                                .append(",,,,")
                                .append(NEW_LINE);

                    }

                    stationSb.append("照准点结束").append(NEW_LINE);
                }
            }

            stationSb.append("END").append(NEW_LINE);
            printSb.append(stationSb);

            printSb.append("[点间数据]");
            printSb.append(NEW_LINE);
            printSb.append("0");
            printSb.append(NEW_LINE);
            printSb.append("END");
            printSb.append(NEW_LINE);
            printSb.append("[导线段数据]");
            printSb.append(NEW_LINE);
            printSb.append("0");
            printSb.append(NEW_LINE);
            printSb.append("END");

            ExportStationMethod.save(printSb.toString(), cycleDir, cycleName + ".kzw");
        }
    }


    @WorkerThread
    private static File exportSingleProject(Context context, Project project, ArrayList<SelectExportType> selectExportTypeList) {
        //读取项目信息
        IProjectDataSource projectDataSource = SurveyorDatabaseFactory.instance.getProjectDataSource();
        ProjectRecord projectRecord = projectDataSource.getRecord(project.projectId);
        if (projectRecord == null) {
            return null;
        }
        //读取限差信息
        String toleranceId = projectRecord.getToleranceId();
        IToleranceDetailDataSource toleranceDataSource = SurveyorDatabaseFactory.instance.getToleranceDetailDataSource();
        ToleranceDetailRecord toleranceRecord = toleranceDataSource.getRecordSync(toleranceId);
        if (toleranceRecord == null) {
            return null;
        }

        //需要导出的project目录
        File projectDir = new File(context.getExternalCacheDir(), projectRecord.getProjectName());
        if (!projectDir.exists()) {
            if (!projectDir.mkdir()) {
                return null;
            }
        }

        if (selectExportTypeList == null) {
            return null;
        }

        for (SelectExportType exportType : selectExportTypeList) {
            if (SelectExportType.TYPE_CONTROL.equals(exportType.getType())) {
                exportBingCe(project, toleranceRecord, projectDir);
            } else if (SelectExportType.TYPE_KE_SHA.equals(exportType.getType())) {
                exportKeSha(project, toleranceRecord, projectDir);
            } else if (SelectExportType.TYPE_NAN_FANG.equals(exportType.getType())) {
                exportSouthAdjustment(project, toleranceRecord, projectDir);
            } else if (SelectExportType.TYPE_QI_FEI.equals(exportType.getType())) {
                exportOpenFly(project, toleranceRecord, projectDir);
            } else if (SelectExportType.TYPE_PING_CHA_ZHI_XING.equals(exportType.getType())) {
                exportPingChaZhiXing(project, toleranceRecord, projectDir);
            }
        }

        ExportStationMethod.save(
                DateUtils.toFull(new Date()) + "\n" + Tools.getVersionName(),
                projectDir,
                ".bc");

        return projectDir;
    }

    @WorkerThread
    private static File exportSingleHeightTraverseProject(Context context, Project project) {
        //读取项目信息
        IProjectDataSource projectDataSource = SurveyorDatabaseFactory.instance.getProjectDataSource();
        ProjectRecord projectRecord = projectDataSource.getRecord(project.projectId);
        if (projectRecord == null) {
            return null;
        }
        //读取限差信息
        String toleranceId = projectRecord.getToleranceId();
        IToleranceDetailDataSource toleranceDataSource = SurveyorDatabaseFactory.instance.getToleranceDetailDataSource();
        ToleranceDetailRecord toleranceRecord = toleranceDataSource.getRecordSync(toleranceId);
        if (toleranceRecord == null) {
            return null;
        }

        //需要导出的project目录
        File projectDir = new File(context.getExternalCacheDir(), projectRecord.getProjectName());
        if (!projectDir.exists()) {
            if (!projectDir.mkdirs()) {
                return null;
            }
        }

        //高程导线
        exportBingceHeightTraverse(project, toleranceRecord, projectDir);
        return projectDir;
    }

    @WorkerThread
    private static void exportBingceHeightTraverse(Project project, ToleranceDetailRecord toleranceRecord, File projectDir) {
        try {
            StringBuilder dxrBuilder = new StringBuilder();
            //遍历每个路线
            for (Map.Entry<String, ExportUtilsPla.SurveyorPeriod> surveyorPeriodEntry : project.surveyorPeriodMap.entrySet()) {
                ExportUtilsPla.SurveyorPeriod surveyorPeriod = surveyorPeriodEntry.getValue();
                ExportStationMethod.exportBingCeSingleStationSurveyorDataHeightTraverse(
                        projectDir,
                        toleranceRecord,
                        project.projectId,
                        surveyorPeriod.periodId, surveyorPeriod.periodName);

                StringBuilder dxrBuilderEveryRoute = ExportStationMethod.exportRouteDxrData(toleranceRecord, project.projectId, surveyorPeriod.periodId);
                dxrBuilder.append(dxrBuilderEveryRoute).append(NEW_LINE);
            }
            //存储dxr
            ExportStationMethod.save(dxrBuilder.toString(), projectDir, project.projectName + ".dxr");
            exportProjectAndControlPoint(project, projectDir);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * 导出平差之星
     */
    private static void exportPingChaZhiXing(Project project, ToleranceDetailRecord toleranceRecord, File projectDir) {
        //读取「限差」信息
        //二级
        ToleranceLevelRecord toleranceLevelRecord = null;
        if (toleranceRecord.levelId != null) {
            toleranceLevelRecord = SurveyorDatabaseFactory.instance.getToleranceLevelDataSource().getRecordSync(toleranceRecord.levelId);
        }
        //一级
        ToleranceGroupRecord toleranceGroupRecord = null;
        if (toleranceRecord.groupId != null) {
            toleranceGroupRecord = SurveyorDatabaseFactory.instance.getToleranceGroupDataSource().getRecordSync(toleranceRecord.groupId);
        }

        //遍历每个测期
        for (Map.Entry<String, SurveyorPeriod> surveyorPeriodEntry : project.surveyorPeriodMap.entrySet()) {
            //初始化天津水平角，垂直角和距离
            StringBuilder horSbTianjinOne = new StringBuilder();
            StringBuilder verSbTianjinOne = new StringBuilder();
            StringBuilder disSbTianjinOne = new StringBuilder();

            SurveyorPeriod surveyorPeriod = surveyorPeriodEntry.getValue();
            //测站文件夹和测站文件的名称暂存map
            Map<String, Integer> stationFileNameMap = new HashMap<>();
            //创建测期文件夹
            File periodFolder = ExportStationMethod.initPeriodFolder(projectDir, surveyorPeriod.periodName);

            for (Map.Entry<String, Station> stationEntry : surveyorPeriod.stationMap.entrySet()) {
                String stationName = stationEntry.getValue().stationName;
                String stationFileName = getStationFileName(stationFileNameMap, stationName);

                ISurveyorStationDataSource stationDataSource = SurveyorDatabaseFactory.instance.getSurveyorStation();
                SurveyorStationRecord stationRecord = stationDataSource.getRecordSync(stationEntry.getValue().stationId);
                if (stationRecord == null) {
                    continue;
                }
                //创建测站文件夹
                File stationFolder = ExportStationMethod.initStationFolder(periodFolder, stationFileName);

                //2.导出单个测站的数据
                ExportStationMethod.exportBingCeSingleStationSurveyorData(
                        stationFolder,
                        toleranceRecord, toleranceLevelRecord, toleranceGroupRecord,
                        stationFileName, stationName, stationRecord);
                //天津收集方向值，垂直角，边长文件的数据
                collectCustomTj61365(stationName, stationRecord, toleranceRecord, horSbTianjinOne, verSbTianjinOne, disSbTianjinOne);
            }

            //输出文件
            if (horSbTianjinOne.length() > 0) {
                horSbTianjinOne.append("END");
                ExportStationMethod.save(horSbTianjinOne.toString(), periodFolder, "方向值.TXT");
            } else {
                ExportStationMethod.save("   ", periodFolder, "方向值.TXT");
            }
            if (verSbTianjinOne.length() > 0) {
                ExportStationMethod.save(verSbTianjinOne.toString(), periodFolder, "垂直角.TXT");
            } else {
                ExportStationMethod.save("   ", periodFolder, "垂直角.TXT");
            }
            if (disSbTianjinOne.length() > 0) {
                disSbTianjinOne.append("END");
                ExportStationMethod.save(disSbTianjinOne.toString(), periodFolder, "边长.TXT");
            } else {
                ExportStationMethod.save("   ", periodFolder, "边长.TXT");
            }
        }
    }

    /**
     * 根据项目列表导出测量数据，所有文件导出到一个zip文件中
     */
    @WorkerThread
    private static File exportAllProject(Context context, Collection<Project> projects, ArrayList<SelectExportType> selectExportTypeList) {
        List<File> zipList = new ArrayList<>();
        for (Project project : projects) {
            File zip = exportSingleProject(context, project, selectExportTypeList);
            if (zip == null) {
                continue;
            }
            zipList.add(zip);
        }
        //将文件存储到一个zip中
        File fileParent = Environment.getExternalStorageDirectory();
        File fileOutput = new File(fileParent, getExportFirstDirectory());
        if (!fileOutput.exists()) {
            fileOutput.mkdir();
        }
//        File fileOutput = PathManager.shareFolder();
        Date date = new Date();
        SimpleDateFormat format = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss", Locale.getDefault());
        String dateString = format.format(date);
        File zip = new File(fileOutput, dateString + ".zip");
        if (!ZipUtil.zip(zip, zipList)) {
            ToastUtils.showShort(R.string.export_failure);
            zip = null;
        }
        //存储成功，删除被压缩的文件
        for (File file : zipList) {
            if (file == null) {
                continue;
            }
            try {
                if (file.isDirectory()) {
                    FileUtils.delete(file);
                } else if (file.isFile()) {
                    file.delete();
                }
            } catch (Exception ignored) {
            }
        }
        return zip;
    }

    /**
     * 根据项目列表导出测量数据，所有文件导出到一个zip文件中
     */
    @WorkerThread
    private static File exportAllHeightTraverseProject(Context context, Collection<Project> projects) {
        List<File> zipList = new ArrayList<>();
        for (Project project : projects) {
            File zip = exportSingleHeightTraverseProject(context, project);
            if (zip == null) {
                continue;
            }
            zipList.add(zip);
        }
        //将文件存储到一个zip中
        File fileParent = Environment.getExternalStorageDirectory();
        File fileOutput = new File(fileParent, getExportFirstDirectory());
        if (!fileOutput.exists()) {
            fileOutput.mkdir();
        }
//        File fileOutput = PathManager.shareFolder();
        Date date = new Date();
        SimpleDateFormat format = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss", Locale.getDefault());
        String dateString = format.format(date);
        File zip = new File(fileOutput, dateString + ".zip");
        if (!ZipUtil.zip(zip, zipList)) {
            ToastUtils.showShort(R.string.export_failure);
            zip = null;
        }
        //存储成功，删除被压缩的文件
        for (File file : zipList) {
            if (file == null) {
                continue;
            }
            try {
                if (file.isDirectory()) {
                    FileUtils.delete(file);
                } else if (file.isFile()) {
                    file.delete();
                }
            } catch (Exception ignored) {
            }
        }
        return zip;
    }

    /**
     * 根据从UI收集的数据，执行导出操作，所有数据导出到一个zip文件中
     */
    @WorkerThread
    public static File doExport(Context context, List<ExportItem> exportItems, ArrayList<SelectExportType> selectExportTypeList) {
        Collection<Project> projects = collect(exportItems);
        return exportAllProject(context, projects, selectExportTypeList);
    }

    /**
     * 根据从UI收集的数据，执行导出操作，所有数据导出到一个zip文件中
     */
    @WorkerThread
    public static File doExportHeightTraverse(Context context, List<ExportItem> exportItems) {
        Collection<Project> projects = collect(exportItems);
        return exportAllHeightTraverseProject(context, projects);
    }

    /**
     * 导线文件导出
     * 一个项目测期的导线对应一个文件
     */
    @WorkerThread
    public static List<File> doExportWire(Context context, List<ExportItem> exportItems, String exportFolder) {
        List<File> exportList = new ArrayList<>();
        if (exportItems == null || exportItems.isEmpty()) {
            return exportList;
        }
        IProjectDataSource projectDataSource = SurveyorDatabaseFactory.instance.getProjectDataSource();
        IToleranceDetailDataSource toleranceDetailDataSource = SurveyorDatabaseFactory.instance.getToleranceDetailDataSource();
        ISurveyorStationDataSource surveyorStationDataSource = SurveyorDatabaseFactory.instance.getSurveyorStation();
        ISurveyorScheduleDataSource surveyorScheduleDataSource = SurveyorDatabaseFactory.instance.getScheduleDataSource();
        IPointDataSource pointDataSource = SurveyorDatabaseFactory.instance.getPointDataSource();
        ILineRecordDataSource lineRecordDatabase = SurveyorDatabaseFactory.instance.getLineRecordDatabase();


        //一个导线一个文件
        boolean isExportSuccess = false;

        for (ExportItem exportItem : exportItems) {

            //校验只能导出导线
            LineRecord lineRecord = lineRecordDatabase.findById(exportItem.stationId);//exportItem.stationId指的的线id
            if (TextUtils.equals(lineRecord.type, LineRecord.TYPE_MUL_ANGLE)) {
                continue;
            }

            //exportItem.stationId这里的id是导线的id 为了复用之前的数据模型
            List<SurveyorStationRecord> stationRecordList = surveyorStationDataSource.getDataListByLine(exportItem.stationId);

            List<PointRecord> unknowPointList = pointDataSource.getListByProjectId(exportItem.projectId, PointRecord.TYPE_UNKNOWN_POINT);
            List<PointRecord> knowPointList = pointDataSource.getListByProjectId(exportItem.projectId, PointRecord.TYPE_KNOWN_POINT);

            StringBuilder printSb = new StringBuilder();
            printSb.append("[PARA]");
            printSb.append(NEW_LINE);

            switch (lineRecord.type) {
                case LineRecord.TYPE_LINE_SINGLE:
                    printSb.append("支导线,0,");
                    break;
                case LineRecord.TYPE_LINE_ATTACH:
                    printSb.append("附合导线,2,");
                    break;
                case LineRecord.TYPE_LINE_CLOSE:
                    printSb.append("闭合导线,1,");
                    break;
                case LineRecord.TYPE_LINE_RANDOM:
                    printSb.append("无定向导线,3,");
                    break;
                default:
                    break;
            }
            printSb.append(unknowPointList.size());
            printSb.append(NEW_LINE);
            printSb.append("[CONTROL]");
            printSb.append(NEW_LINE);

            for (int i = 0; i < knowPointList.size(); i++) {
                PointRecord knowPoint = knowPointList.get(i);
                printSb.append(knowPoint.name + ","
                        + ExportStationMethod.getXyzFormat(knowPoint.getX()) + ","
                        + ExportStationMethod.getXyzFormat(knowPoint.getY()) + ","
                        + ExportStationMethod.getXyzFormat(knowPoint.getZ()));
                printSb.append(NEW_LINE);
            }

            printSb.append("[OBSER]");
            printSb.append(NEW_LINE);

            //遍历测站
            for (int i = 0; i < stationRecordList.size(); i++) {
                SurveyorStationRecord stationRecord = stationRecordList.get(i);

                List<SurveyorPoint> items = stationRecord.getItems();

                double stationDeviceHeight = SurveyorPoint.getDoubleValueFromString(stationRecord.instrumentHeight);
                SurveyorPoint lookPoint = items.get(items.size() - 1);
                double prismHeight = SurveyorPoint.getDoubleValueFromString(lookPoint.getPrismHeight());

                PointRecord pointRecord = pointDataSource.findByIdSync(items.get(0).getPointId());


                ProjectRecord projectRecord = projectDataSource.getRecord(exportItem.projectId);
                ToleranceDetailRecord toleranceRecord = toleranceDetailDataSource.getRecordSync(projectRecord.getToleranceId());

                //创建缓存数据
                CachedData cachedData =
                        CachedData.newInstance(stationRecord.isAutoClose(),
                                stationRecord.isLeftAndRightSurveyor(),
                                stationRecord.getId(), toleranceRecord);


                final double[] horAngle = new double[1];
                final double[] verAngle = new double[1];
                final double[] distance = new double[1];

                SurveyorAdjustMethod.AdjustResult adjustResult = SurveyorAdjustMethod.calStationAdjust(stationRecord, cachedData);
                if (TextUtils.isEmpty(adjustResult.errorString)) {
                    //展示结果数据
                    //检测归零值是否为负，如果是需要+2PI
                    for (SurveyorAdjustMethod.Item item : adjustResult.result) {
                        if (item.getHorAngle() < 0) {
                            item.setHorAngle(Math.PI * 2 + item.getHorAngle());
                        }
                    }
                    SurveyorAdjustMethod.Item item = adjustResult.result.get(adjustResult.result.size() - 1);
                    horAngle[0] = item.getHorAngle();
                    verAngle[0] = item.getVerAngle();
                    distance[0] = item.getDistance();
                }

                //水平角 保留5位小数
                String horAngleOutPut = Util.formatDouble2String(Util.radianToDmsDouble(horAngle[0]), ANGLE_DECIMALS);

                printSb
                        .append(pointRecord.getName())
                        .append(",");


                if (i == 0 || i != stationRecordList.size() - 1) {
                    //平距
                    double horDistance = getHorDistance(verAngle[0], distance[0]);
                    String horDistanceOutput = Util.formatDouble2String(horDistance, ANGLE_DECIMALS);
                    //高差
                    double altitudeDifference = getAltitudeDifference(verAngle[0], distance[0], stationDeviceHeight, prismHeight);
                    String altitudeDifferenceOutput = Util.formatDouble2String(altitudeDifference, ANGLE_DECIMALS);

                    appendHorAngle(printSb, horAngleOutPut);
                    printSb
                            .append(",")
                            .append(horDistanceOutput) //斜距distance[0] 改为平距
                            .append(",")
                            .append(altitudeDifferenceOutput) //高差
                            .append(",")
                            .append(horDistanceOutput); //平距
                } else {
                    //最后一个
                    if (TextUtils.equals(lineRecord.type, LineRecord.TYPE_LINE_CLOSE)
                            || TextUtils.equals(lineRecord.type, LineRecord.TYPE_LINE_ATTACH)) {
                        appendHorAngle(printSb, horAngleOutPut);
                    }
                    printSb.append(",,,");
                }

                printSb.append(NEW_LINE);

                if (TextUtils.equals(lineRecord.type, LineRecord.TYPE_LINE_SINGLE)) {
                    PointRecord pointRecordLast = pointDataSource.findByIdSync(items.get(2).getPointId());
                    printSb.append(pointRecordLast.getName())
                            .append(",,,,");
                    printSb.append(NEW_LINE);
                }


            }

            printSb.append("[END]");

            String absolutePath = Environment.getExternalStorageDirectory().getAbsolutePath();
            File fileDir = new File(absolutePath, exportFolder);

            String exportFirstName = exportItem.projectName + "_" + exportItem.cycleName + "_" + exportItem.stationName;

            String fileName = exportFirstName + ".dxp";
            fileName = fileName.replaceAll(":", "_");
//            FileUtil.writeString(printSb.toString(), new File(FileUtil.getSDPath()+"/234/"+fileName));

            ExportStationMethod.save(printSb.toString(), fileDir, fileName);

            File fileHint = new File(fileDir, fileName);
            if (fileHint.exists() && fileHint.length() > 0) {
                //一个文件导出成功就算成功了
                exportList.add(fileHint);
                isExportSuccess = true;
            }
        }

        return exportList;

    }

    /**
     * 添加水平角
     *
     * @param sb
     * @param horAngleOutPut
     */
    private static void appendHorAngle(StringBuilder sb, String horAngleOutPut) {
        if (!TextUtils.isEmpty(horAngleOutPut) && !TextUtils.equals("0.00000", horAngleOutPut)) {
            sb.append(horAngleOutPut);
        }
    }

    /**
     * 返回平距
     *
     * @param verAngle
     * @param distance
     * @return
     */
    private static double getHorDistance(double verAngle, double distance) {
        return Math.abs(distance * Math.cos(Math.PI / 2 - verAngle));
    }

    /**
     * 返回高差
     *
     * @param verAngle
     * @param distance
     * @param stationDeviceHeight
     * @param prismHeight
     * @return
     */
    private static double getAltitudeDifference(double verAngle, double distance, double stationDeviceHeight, double prismHeight) {
        return distance * Math.sin(Math.PI / 2 - verAngle) - prismHeight + stationDeviceHeight;
    }

    static class Station {
        final String stationId;
        final String stationName;

        Station(String id, String name) {
            stationId = id;
            stationName = name;
        }
    }

    static class SurveyorPeriod {
        final String periodId;
        final String periodName;
        final Map<String, Station> stationMap = new LinkedHashMap<>();//key stationId

        public SurveyorPeriod(String id, String name) {
            periodId = id;
            periodName = name;
        }
    }

    static class Project {
        final String projectId;
        final String projectType;
        final String projectName;
        final Map<String, SurveyorPeriod> surveyorPeriodMap = new LinkedHashMap<>();//key 测期id

        public Project(String id, String projectType, String name) {
            projectId = id;
            this.projectType = projectType;
            projectName = name;
        }
    }

    public static class ExportItem {
        final String projectName;
        final String projectId;
        final String projectType;
        final String cycleId;
        final String cycleName;
        final String stationId;
        final String stationName;
        final String stationPointName;

        public ExportItem(String projectName, String projectId, String projectType,
                          String cycleId, String cycleName,
                          String stationId, String stationName, String stationPointName) {
            this.projectName = projectName;
            this.projectId = projectId;
            this.projectType = projectType;
            this.cycleId = cycleId;
            this.cycleName = cycleName;
            this.stationId = stationId;
            this.stationName = stationName;
            this.stationPointName = stationPointName;
        }

        @Override
        public @NotNull String toString() {
            return stationName;
        }
    }

    /**
     * 获取天津导出的项目类型
     *
     * @param projectType
     * @return 0 三角 1 联系 3 高程 4 水准
     */
    public static int getProjectTypeForTianjin(String projectType) {
        switch (projectType) {
            case ProjectTypeValue.TRIANGLE:
                return 0;
            case ProjectTypeValue.CONNECTION:
                return 1;
            case ProjectTypeValue.HEIGHT_TRAVERSE:
                return 2;
            case ProjectTypeValue.LEVEL:
                return 3;
            default:
                break;
        }
        return -1;
    }

    /**
     * 获取天津导出的点类型
     *
     * @param pointType
     * @return 0 已知点 1 未知点
     */
    public static int getPointTypeForTianjin(String pointType) {
        switch (pointType) {
            case PointRecord.TYPE_KNOWN_POINT:
            case PointRecord.TYPE_GLOBAL_POINT:
                return ExportContants.TYPE_KNOWN_POINT;
            case PointRecord.TYPE_UNKNOWN_POINT:
                return ExportContants.TYPE_UNKNOWN_POINT;
            default:
                break;
        }
        return -1;
    }

    /**
     * 格式化仪器高或者棱镜高
     *
     * @param height
     * @return
     */
    private static String formatHeight(double height) {
        return Tools.formatDouble2String(height, 3);
    }

    public static boolean isCanExport() {
        if (CheckRegUtil.checkReg(App.getApp().registerCodeV2)) {
            return true;
        }
        ToastUtils.showShort("您暂无使用权限，请联系客服");
        return false;
    }

    /**
     * 格式化仪器高或者棱镜高
     *
     * @param distance
     * @return
     */
    public static String formatDistance(double distance) {
        return Tools.formatDouble2String(distance, DISTANCE_DECIMALS);
    }

    /**
     * 格式化ppm
     *
     * @param distance
     * @return
     */
    public static String formatPPM(double distance) {
        return Tools.formatDouble2String(distance, PPM_DECIMALS);
    }


}
