package com.bingce.controlnetwork.surveyor.state.angledistance;

import android.os.Bundle;

import androidx.annotation.WorkerThread;
import androidx.fragment.app.FragmentManager;
import androidx.lifecycle.LifecycleOwner;

import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory;
import com.bingce.controlapphelper.datasource.database.point.IPointDataSource;
import com.bingce.controlapphelper.datasource.database.point.PointRecord;
import com.bingce.controlapphelper.datasource.database.schedule.SurveyorScheduleConstants;
import com.bingce.controlapphelper.datasource.database.surveyorstation.ISurveyorStationDataSource;
import com.bingce.controlapphelper.datasource.database.surveyorstation.SurveyorStationRecord;
import com.bingce.controlapphelper.datasource.database.surveyorstation.model.SurveyorPoint;
import com.bingce.controlapphelper.surveyor.data.ICachedData;
import com.bingce.controlapphelper.surveyor.data.ICycleRecorder;
import com.bingce.controlapphelper.surveyor.data.ISurveyorData;
import com.bingce.controlapphelper.surveyor.state.StateType;
import com.bingce.controlapphelper.util.Tools;
import com.bingce.controlnetwork.R;
import com.bingce.controlnetwork.fragment.SelectReSurveyorPointFragment;
import com.bingce.controlnetwork.surveyor.util.checkadjust.AngleDistanceCheckAdjust;
import com.bingce.controlnetwork.surveyor.util.checkadjust.IStationCheck;
import com.bingce.utils.ThreadPoolUtil;

import java.util.ArrayList;
import java.util.List;

/**
 * 执行「测站检测」
 * <p>
 * 前置条件：
 * 所有测回都完成测量，并且通过「测回内检测」
 */
class _CheckStationState extends __BaseState {

    private final IStationCheck stationCheck = new AngleDistanceCheckAdjust();

    public _CheckStationState(StateManager stateManager) {
        super(StateType.checkStation, stateManager);
    }

    @Override
    public void enter(Bundle args) {
        //执行测站检测
        String error = stationCheck.check(stationId(), getToleranceRecord(), getCachedData());
        if (error == null || error.isEmpty()) {
            //进入finish状态
            navigation2(StateType.stationCheckedSuccess);
        } else {
            showErrorTips(error, stationId(), getToleranceRecord().getCycleCount(),
                    getSurveyorData(), getCycleRecorder(), getCachedData(),
                    this,
                    fragmentManager(),
                    lifecycleOwner());
        }
    }

    /**
     * 提示错误信息，并让用户选择下一步操作
     */
    static void showErrorTips(String msg, String stationId, int cycleCount,
                              ISurveyorData surveyorDataSource,
                              ICycleRecorder cycleRecorder,
                              ICachedData stationSurveyorDataSource,
                              __BaseState state,
                              FragmentManager fragmentManager, LifecycleOwner lifecycleOwner) {
        state.optionTips(msg,
                Tools.getString(R.string.retest_direction),
                Tools.getString(R.string.retest_station),
                () -> state.runOnBackground(() -> reSurveyorPoint(stationId, cycleCount,
                        cycleRecorder, stationSurveyorDataSource,
                        state, fragmentManager, lifecycleOwner)),
                () -> reSurveyorStation(msg, stationId, cycleCount,
                        surveyorDataSource, cycleRecorder, stationSurveyorDataSource,
                        state, fragmentManager, lifecycleOwner),
                () -> state.enterIdleAndClearSurveyStateAndResetAutoMode(true));
    }

    @WorkerThread
    static void reSurveyorPoint(String stationId, int cycleCount,
                                ICycleRecorder cycleRecorder,
                                ICachedData cachedData,
                                __BaseState state,
                                FragmentManager fragmentManager, LifecycleOwner lifecycleOwner) {
        //弹框提示选择重测数据
        ISurveyorStationDataSource dataSource = SurveyorDatabaseFactory.instance.getSurveyorStation();
        SurveyorStationRecord instance = dataSource.getRecordSync(stationId);
        if (instance == null) {
            state.enterIdleAndClearSurveyStateAndResetAutoMode(true);
            return;
        }
        IPointDataSource pointDataSource = SurveyorDatabaseFactory.instance.getPointDataSource();
        List<SurveyorPoint> stationItems = instance.getItems();
        ArrayList<String> ids = new ArrayList<>();
        ArrayList<String> names = new ArrayList<>();
        for (SurveyorPoint item : stationItems) {
            if (item.isAbandon()) {
                continue;
            }

            if (!SurveyorPoint.TYPE_POINT.equals(item.getType())) {
                continue;
            }
            PointRecord pointRecord = pointDataSource.findByIdSync(item.getPointId());
            if (pointRecord == null) {
                continue;
            }
            ids.add(pointRecord.id);
            names.add(pointRecord.name);
        }
        state.runOnUI(() -> SelectReSurveyorPointFragment.pick(
                fragmentManager,
                lifecycleOwner,
                cycleCount,
                instance.leftAndRightSurveyor,
                ids,
                names,
                (cycleIndex, pointIds) -> {
                    if (pointIds == null || pointIds.isEmpty()) {
                        state.enterIdleAndClearSurveyStateAndResetAutoMode(true);
                    } else {
                        //记录补测数据类型，切换测回
                        cachedData.markReSurveyor(stationId, cycleIndex, pointIds);
                        cycleRecorder.changeCycleIndex(cycleIndex);

                        //更新测量状态
                        state.setCurrentCycle(cycleIndex);
                        state.setAllCycleSchedule(SurveyorScheduleConstants.STATUE_VALUE_NO_COMPLETED);
                        state.setSingleCycleSchedule(cycleIndex, SurveyorScheduleConstants.STATUE_VALUE_NO_COMPLETED);

                        //查找新测量目标
                        state.findTargetFromBeginInCurrentCycle(true);
                    }
                }
        ));
    }

    static void reSurveyorStation(String msg, String stationId, int cycleCount,
                                  ISurveyorData surveyorData,
                                  ICycleRecorder cycleRecorder,
                                  ICachedData cachedData,
                                  __BaseState state,
                                  FragmentManager fragmentManager, LifecycleOwner lifecycleOwner) {
        state.optionTips(Tools.getString(R.string.is_delete_all_station_data), null, null,
                () -> {//1.选择删除测站所有数据，则执行相应逻辑
                    ThreadPoolUtil.execute(() -> {
                        //删除数据
                        cachedData.deleteSurveyorRecord(stationId);

                        //更新测量进度
                        state.setAllCycleSchedule(SurveyorScheduleConstants.STATUE_VALUE_NO_COMPLETED);
                        for (int index = 0; index < cycleCount; index++) {
                            state.setSingleCycleSchedule(index, SurveyorScheduleConstants.STATUE_VALUE_NO_COMPLETED);
                        }

                        state.resetLearnDataTag();

                        //重新测量
                        state.findTargetFromBeginInCurrentCycle(true);
                    });
                },
                () -> {//2.如果用户不选择删除，则循环提示，让他重新在重测和删除做选择
                    //循环显示
                    showErrorTips(msg, stationId, cycleCount,
                            surveyorData, cycleRecorder, cachedData, state,
                            fragmentManager, lifecycleOwner);
                },
                () -> state.enterIdleAndClearSurveyStateAndResetAutoMode(true));//3.如果用户忽略，则停留在当前位置
    }

}
