package com.bingce.controlnetwork.surveyor.state.heighttraverse;

import android.os.Bundle;

import androidx.annotation.WorkerThread;
import androidx.fragment.app.FragmentManager;
import androidx.lifecycle.LifecycleOwner;

import com.bingce.controlapphelper.datasource.database.SurveyorDatabaseFactory;
import com.bingce.controlapphelper.datasource.database.schedule.SurveyorScheduleConstants;
import com.bingce.controlapphelper.datasource.database.surveyorstation.SurveyorStationRecord;
import com.bingce.controlapphelper.datasource.database.tolerance.detail.ToleranceDetailRecord;
import com.bingce.controlapphelper.model.LeftOrRight;
import com.bingce.controlapphelper.surveyor.data.ICachedData;
import com.bingce.controlapphelper.surveyor.method.AverageValueMethod;
import com.bingce.controlapphelper.surveyor.method.DeviationValueOfRadianMethod;
import com.bingce.controlapphelper.surveyor.method.RadianMethod;
import com.bingce.controlapphelper.surveyor.method.ValueOf2cMethod;
import com.bingce.controlapphelper.surveyor.method.model.CacheResult;
import com.bingce.controlapphelper.surveyor.method.model.SdAngleValueType;
import com.bingce.controlapphelper.surveyor.state.StateType;
import com.bingce.controlapphelper.util.IdUtil;
import com.bingce.controlapphelper.util.SurveyorCycleUtil;
import com.bingce.controlapphelper.util.Tools;
import com.bingce.controlnetwork.R;
import com.bingce.controlnetwork.fragment.SelectReSurveyorCycleFragment;
import com.bingce.utils.ThreadPoolUtil;
import com.bingce.utils.Util;

import java.util.ArrayList;
import java.util.List;

/**
 * 执行「测站检测」
 * <p>
 * 前置条件：
 * 所有测回都完成测量，并且通过「测回内检测」
 */
class _CheckStationState extends __BaseState {

    public _CheckStationState(int positionFragment, StateManagerHeightTraverse stateManager) {
        super(StateType.checkStation, positionFragment, stateManager);
    }

    @Override
    public void enter(Bundle args) {
        //执行测站检测
        String error = _stationCheck(stationId(), getToleranceRecord(), getCachedData());
        if (error.isEmpty()) {
            //进入finish状态
            navigation2(StateType.stationCheckedSuccess);
        } else {
            showErrorTips(error, stationId(), getToleranceRecord().getCycleCount(),
                    getCachedData(),
                    this,
                    fragmentManager(),
                    lifecycleOwner());
        }
    }

    /**
     * 提示错误信息，并让用户选择下一步操作
     */
    public static void showErrorTips(String msg, String stationId, int cycleCount,
                                     ICachedData stationSurveyorDataSource,
                                     __BaseState state,
                                     FragmentManager fragmentManager, LifecycleOwner lifecycleOwner) {
        state.optionTips(msg,
                Tools.getString(R.string.retest_cycle),
                Tools.getString(R.string.retest_station),
                () -> state.runOnBackground(() -> reSurveyorCycle(stationId, cycleCount,
                        stationSurveyorDataSource,
                        state, fragmentManager, lifecycleOwner)),
                () -> reSurveyorStation(msg, stationId, cycleCount,
                        stationSurveyorDataSource,
                        state, fragmentManager, lifecycleOwner),
                () -> state.enterIdleAndClearSurveyStateAndResetAutoMode(true));
    }


    @WorkerThread
    private static void reSurveyorCycle(String stationId, int cycleCount,
                                        ICachedData cachedData,
                                        __BaseState state,
                                        FragmentManager fragmentManager, LifecycleOwner lifecycleOwner) {
        SurveyorStationRecord stationRecord = SurveyorDatabaseFactory.instance.getSurveyorStation().getRecordSync(stationId);
        if (stationRecord == null) {
            state.enterIdleAndClearSurveyStateAndResetAutoMode(true);
            return;
        }
        state.runOnUI(() -> SelectReSurveyorCycleFragment.start(
                fragmentManager,
                lifecycleOwner,
                cycleCount,
                (cycleIndex) -> {
                    if (cycleIndex == -1) {
                        state.enterIdleAndClearSurveyStateAndResetAutoMode(true);
                    } else {
                        //记录补测数据类型
                        List<String> pointIds = new ArrayList<>();
                        pointIds.add(cachedData.pointId(cycleIndex, cycleIndex));
                        cachedData.markReSurveyor(stationId, cycleIndex, pointIds);
                        //更新测量状态
                        state.setCurrentCycle(cycleIndex);
                        state.setAllCycleSchedule(SurveyorScheduleConstants.STATUE_VALUE_NO_COMPLETED);
                        state.setSingleCycleSchedule(cycleIndex, SurveyorScheduleConstants.STATUE_VALUE_NO_COMPLETED);

                        //查找新测量目标
                        state.findTargetFromBeginInCurrentCycle(true, cycleIndex);
                    }
                }
        ));
    }

    static void reSurveyorStation(String msg, String stationId, int cycleCount,
                                  ICachedData cachedData,
                                  __BaseState state,
                                  FragmentManager fragmentManager, LifecycleOwner lifecycleOwner) {
        state.optionTips(Tools.getString(R.string.is_delete_all_station_data), null, null,
                () -> {//1.选择删除测站所有数据，则执行相应逻辑
                    ThreadPoolUtil.execute(() -> {
                        //删除数据
                        cachedData.deleteSurveyorRecord(stationId);
                        //更新测量进度
                        state.setAllCycleSchedule(SurveyorScheduleConstants.STATUE_VALUE_NO_COMPLETED);
                        for (int index = 0; index < cycleCount; index++) {
                            state.setSingleCycleSchedule(index, SurveyorScheduleConstants.STATUE_VALUE_NO_COMPLETED);
                        }

                        state.resetLearnDataTag();

                        //重新测量
                        state.findTargetFromBeginInCurrentCycle(true, 0);
                    });
                },
                () -> {//2.如果用户不选择删除，则循环提示，让他重新在重测和删除做选择
                    //循环显示
                    showErrorTips(msg, stationId, cycleCount,
                            cachedData, state,
                            fragmentManager, lifecycleOwner);
                },
                () -> state.enterIdleAndClearSurveyStateAndResetAutoMode(true));//3.如果用户忽略，则停留在当前位置
    }


    //检测合法性

    /**
     * 检测所有测回数据
     * <p>
     * 在「无自动闭合」情况下，为了兼容「左右角」测量结果，根据pointId将所有测量数据存入map中
     * 根据「左角」测量来检测「测回间同方向值互差」，所有测回的归零值使用「左角」第一个point来做0点
     * <p>
     * 如果是「自动闭合」情况，则忽略「左右角」测量，直接按照itemIndex依次计算，第个测回的归零值，
     * 使用每次循环的itemIndex=0做0点
     *
     * @param toleranceRecord  限差
     * @param cachedDataSource 缓存的测量数据
     * @return 错误信息，如果为null或空，说明没有错误
     */
    public static String _stationCheck(final String stationId,
                                       final ToleranceDetailRecord toleranceRecord,
                                       final ICachedData cachedDataSource) {
        final int cycleCount = toleranceRecord.getCycleCount();
        final int surveyorCount = toleranceRecord.getSurveyorCount();

        //*****测回间垂直角互差（垂直角）
        double verCyclesAngleDifference = toleranceRecord.getVerCyclesAngleDifference()/*verCyclesAngleDifference*/;
        List<_ErrorItem> verValueError = new ArrayList<>();

        //执行计算
        for (int cycleI = 0; cycleI < cycleCount; cycleI++) {
            List<DeviationValueOfRadianMethod.AngleDistValue> valuesI =
                    DeviationValueOfRadianMethod.zeroValueOfSingleCycleInLeftOrientation(stationId, cycleI, surveyorCount, cachedDataSource);
            for (int cycleJ = cycleI + 1; cycleJ < cycleCount; cycleJ++) {
                List<DeviationValueOfRadianMethod.AngleDistValue> valuesJ =
                        DeviationValueOfRadianMethod.zeroValueOfSingleCycleInLeftOrientation(stationId, cycleJ, surveyorCount, cachedDataSource);
                if (valuesI.size() != valuesJ.size()) {
                    return Tools.getString(R.string.please_complete_all_retest_data);
                }
                int size = valuesI.size();
                for (int index = 0; index < 1; index++) {
                    double diff = Math.abs(DeviationValueOfRadianMethod
                            .similarRadianDiff(valuesI.get(index).verRadian, valuesJ.get(index).verRadian));
                    if (verCyclesAngleDifference < RadianMethod.radian2S(diff)) {
                        verValueError.add(new _ErrorItem(valuesI.get(index).pointName,
                                cycleI, cycleJ, RadianMethod.radian2S(diff)));
                    }
                }
            }
        }
        //*****测回间垂直角I互差（垂直角）
        double verCyclesIDifference = toleranceRecord.getVerCyclesIDifference()/*verCyclesIDifference*/;
        List<_ErrorItem> verIError = new ArrayList<>();
        //*****测回间斜距互差（距离）
        double distCyclesDifference = toleranceRecord.getSlopeCycleDistanceDifference()/*slopeCycleDistanceDifference*/;
        List<_ErrorItem> distError = new ArrayList<>();

        List<_Point> pointArrayList = new ArrayList<>();
        //遍历第一测回，收集所有待测点id
        for (int index = 0; index < 1; index++) {
            if (cachedDataSource.isPointAbandon(0, index)) {
                continue;
            }
            pointArrayList.add(new _Point(cachedDataSource.pointId(0, index),
                    cachedDataSource.pointName(0, index),
                    cachedDataSource.isAloneDistance(0, index)));
        }
        //检测所有待测点i互差
        for (_Point point : pointArrayList) {
            for (int cycleI = 0; cycleI < cycleCount; cycleI++) {
                String leftRecordId_I = IdUtil.createSurveyorRecordId(stationId, cycleI, point.pointId, LeftOrRight.left);
                String rightRecordId_I = IdUtil.createSurveyorRecordId(stationId, cycleI, point.pointId, LeftOrRight.right);
                CacheResult value2C_I = ValueOf2cMethod.valueOf2C(SdAngleValueType.verAngle, leftRecordId_I, rightRecordId_I,
                        toleranceRecord.getSurveyorCount(), cachedDataSource);
                CacheResult distanceLeftAndRightAverageValue_I = AverageValueMethod.leftAndRightValue(SdAngleValueType.distance,
                        leftRecordId_I,
                        rightRecordId_I, surveyorCount, cachedDataSource);
                for (int cycleJ = cycleI + 1; cycleJ < cycleCount; cycleJ++) {
                    String leftRecordId_J = IdUtil.createSurveyorRecordId(stationId, cycleJ, point.pointId, LeftOrRight.left);
                    String rightRecordId_J = IdUtil.createSurveyorRecordId(stationId, cycleJ, point.pointId, LeftOrRight.right);
                    CacheResult value2C_J = ValueOf2cMethod.valueOf2C(SdAngleValueType.verAngle, leftRecordId_J, rightRecordId_J,
                            toleranceRecord.getSurveyorCount(), cachedDataSource);
                    CacheResult distanceLeftAndRightAverageValue_J = AverageValueMethod.leftAndRightValue(SdAngleValueType.distance,
                            leftRecordId_J,
                            rightRecordId_J, surveyorCount, cachedDataSource);
                    double verI = Math.abs(value2C_I.value - value2C_J.value);
                    if (verCyclesIDifference < verI) {
                        verIError.add(new _ErrorItem(point.pointName, cycleI, cycleJ, verI));
                    }

                    if (!point.aloneDistance) {
                        double distance = Math.abs(distanceLeftAndRightAverageValue_I.value - distanceLeftAndRightAverageValue_J.value);
                        if (distCyclesDifference < distance * 1000) {
                            distError.add(new _ErrorItem(point.pointName, cycleI, cycleJ, distance * 1000));
                        }
                    }

                }
            }
        }

        return _errorString(verCyclesAngleDifference, verCyclesIDifference, distCyclesDifference,
                verValueError, verIError, distError);
    }

    /**
     * 根据超限信息生成错误描述
     *
     * @param verValueError 垂直角同方向值互差
     */
    private static String _errorString(double verLimit, double verILimit, double distLimit
            , List<_ErrorItem> verValueError, List<_ErrorItem> verIError, List<_ErrorItem> distError) {
        StringBuilder stringBuilder = new StringBuilder();
        if (!verValueError.isEmpty()) {
            stringBuilder.append("\n").append(Tools.getString(R.string.cycle_between_ver_angle_limit))
                    .append(Util.formatDouble2String(verLimit, 1))
                    .append("'')");
            for (_ErrorItem item : verValueError) {
                stringBuilder
                        .append("\n   ")
                        .append(item.pointName)
                        .append(":\n")
                        .append("   |")
                        .append(SurveyorCycleUtil.cycleName(item.cycleIndex1))
                        .append("-")
                        .append(SurveyorCycleUtil.cycleName(item.cycleIndex2))
                        .append("| = ")
                        .append(Util.formatDouble2String(item.difference, 1))
                        .append("'')");
            }
        }

        if (!verIError.isEmpty()) {
            stringBuilder.append("\n").append(Tools.getString(R.string.ver_angle_i_same_direction_each_cycle))
                    .append(Util.formatDouble2String(verILimit, 1))
                    .append("'')");
            for (_ErrorItem item : verIError) {
                stringBuilder
                        .append("\n   ")
                        .append(item.pointName)
                        .append(":\n")
                        .append("   |")
                        .append(SurveyorCycleUtil.cycleName(item.cycleIndex1))
                        .append("-")
                        .append(SurveyorCycleUtil.cycleName(item.cycleIndex2))
                        .append("| = ")
                        .append(Util.formatDouble2String(item.difference, 1))
                        .append("'')");
            }
        }

        if (!distError.isEmpty()) {
            stringBuilder.append("\n").append(Tools.getString(R.string.cycle_between_distance_limit_brackets))
                    .append(Util.formatDouble2String(distLimit, 1))
                    .append("'')");
            for (_ErrorItem item : distError) {
                stringBuilder
                        .append("\n   ")
                        .append(item.pointName)
                        .append(":\n")
                        .append("   |")
                        .append(SurveyorCycleUtil.cycleName(item.cycleIndex1))
                        .append("-")
                        .append(SurveyorCycleUtil.cycleName(item.cycleIndex2))
                        .append("| = ")
                        .append(Util.formatDouble2String(item.difference, 1))
                        .append("'')");
            }
        }

        return stringBuilder.toString();
    }

    /**
     * 存储错误信息的item
     */
    private static class _ErrorItem {
        final String pointName;
        final int cycleIndex1;
        final int cycleIndex2;
        final double difference;

        public _ErrorItem(String pointName, int cycleIndex1, int cycleIndex2, double difference) {
            this.pointName = pointName;
            this.cycleIndex1 = cycleIndex1;
            this.cycleIndex2 = cycleIndex2;
            this.difference = difference;
        }
    }

    private static class _Point {
        final String pointId;
        final String pointName;
        final boolean aloneDistance;

        _Point(String pointId, String pointName, boolean aloneDistance) {
            this.pointId = pointId;
            this.pointName = pointName;
            this.aloneDistance = aloneDistance;
        }
    }
}
