package com.bingce.controlnetwork.surveyor.util.checkadjust

import com.bingce.controlapphelper.datasource.database.tolerance.detail.ToleranceDetailRecord
import com.bingce.controlapphelper.model.LeftOrRight
import com.bingce.controlapphelper.surveyor.data.ICachedData
import com.bingce.controlapphelper.surveyor.data.IMeasureCache
import com.bingce.controlapphelper.surveyor.method.AverageValueMethod
import com.bingce.controlapphelper.surveyor.method.DeviationValueOfRadianMethod
import com.bingce.controlapphelper.surveyor.method.RadianMethod
import com.bingce.controlapphelper.surveyor.method.ValueOf2cMethod
import com.bingce.controlapphelper.surveyor.method.model.SdAngleValueType
import com.bingce.controlapphelper.util.IdUtil
import com.bingce.controlapphelper.util.SurveyorCycleUtil
import com.bingce.utils.Util
import kotlin.math.abs

/**
 * 常规三角导线测站检测
 * 检测所有测回数据
 * <p>
 * 在「无自动闭合」情况下，为了兼容「左右角」测量结果，根据pointId将所有测量数据存入map中
 * 根据「左角」测量来检测「测回间同方向值互差」，所有测回的归零值使用「左角」第一个point来做0点
 * <p>
 * 如果是「自动闭合」情况，则忽略「左右角」测量，直接按照itemIndex依次计算，第个测回的归零值，
 * 使用每次循环的itemIndex=0做0点
 *
 * @param toleranceRecord  限差
 * @param cachedDataSource 缓存的测量数据
 * @return 错误信息，如果为null或空，说明没有错误
 */
class AngleDistanceCheckAdjust : IStationCheck {
    override fun check(
        stationId: String,
        toleranceRecord: ToleranceDetailRecord,
        measureCache: IMeasureCache
    ): String? {
        val cacheData = measureCache as ICachedData

        val cycleCount = toleranceRecord.getCycleCount()
        val surveyorCount = toleranceRecord.surveyorCount

        //*****测回间同方向值互差(水平角)
        val horCyclesDirectValueDifference =
            toleranceRecord.getHorCyclesDirectValueDifference() /*horCyclesDirectValueDifference*/
        val horValueError: MutableList<_ErrorItem> = ArrayList()

        //*****测回间垂直角互差（垂直角）
        val verCyclesAngleDifference =
            toleranceRecord.getVerCyclesAngleDifference() /*verCyclesAngleDifference*/
        val verValueError: MutableList<_ErrorItem> = ArrayList()


        //执行计算
        for (cycleI in 0 until cycleCount) {
            val valuesI =
                DeviationValueOfRadianMethod.zeroValueOfSingleCycleInLeftOrientation(
                    stationId,
                    cycleI,
                    surveyorCount,
                    cacheData
                )
            for (cycleJ in cycleI + 1 until cycleCount) {
                val valuesJ =
                    DeviationValueOfRadianMethod.zeroValueOfSingleCycleInLeftOrientation(
                        stationId,
                        cycleJ,
                        surveyorCount,
                        cacheData
                    )
                if (valuesI.size != valuesJ.size) {
                    return "请完成所有重测后再检测测站数据！"
                }
                val size = valuesI.size
                for (index in 0 until size) {
                    var diff = abs(
                        DeviationValueOfRadianMethod
                            .similarRadianDiff(
                                valuesI[index].horRadian,
                                valuesJ[index].horRadian
                            )
                    )
                    if (horCyclesDirectValueDifference < RadianMethod.radian2S(diff)) {
                        horValueError.add(
                            _ErrorItem(
                                valuesI[index].pointName,
                                cycleI, cycleJ, RadianMethod.radian2S(diff)
                            )
                        )
                    }

                    diff = abs(
                        DeviationValueOfRadianMethod
                            .similarRadianDiff(valuesI[index].verRadian, valuesJ[index].verRadian)
                    )
                    if (verCyclesAngleDifference < RadianMethod.radian2S(diff)) {
                        verValueError.add(
                            _ErrorItem(
                                valuesI[index].pointName,
                                cycleI, cycleJ, RadianMethod.radian2S(diff)
                            )
                        )
                    }
                }
            }
        }

        //*****测回间垂直角I互差（垂直角）
        val verCyclesIDifference =
            toleranceRecord.getVerCyclesIDifference() /*verCyclesIDifference*/
        val verIError: MutableList<_ErrorItem> = ArrayList()

        //*****测回间斜距互差（距离）
        val distCyclesDifference =
            toleranceRecord.getSlopeCycleDistanceDifference() /*slopeCycleDistanceDifference*/
        val distError: MutableList<_ErrorItem> = ArrayList()

        val pointArrayList: MutableList<_Point> = ArrayList()

        //遍历第一测回，收集所有待测点id
        for (index in 0 until cacheData.pointCount()) {
            if (cacheData.isPointAbandon(0, index)) {
                continue
            }
            pointArrayList.add(
                _Point(
                    cacheData.pointId(0, index),
                    cacheData.pointName(0, index),
                    cacheData.isAloneDistance(0, index)
                )
            )
        }

        //检测所有待测点i互差
        for (point in pointArrayList) {
            for (cycleI in 0 until cycleCount) {
                val leftRecordId_I = IdUtil.createSurveyorRecordId(
                    stationId,
                    cycleI,
                    point.pointId,
                    LeftOrRight.left
                )
                val rightRecordId_I = IdUtil.createSurveyorRecordId(
                    stationId,
                    cycleI,
                    point.pointId,
                    LeftOrRight.right
                )
                val value2C_I = ValueOf2cMethod.valueOf2C(
                    SdAngleValueType.verAngle, leftRecordId_I, rightRecordId_I,
                    toleranceRecord.surveyorCount, cacheData
                )
                val distanceLeftAndRightAverageValue_I = AverageValueMethod.leftAndRightValue(
                    SdAngleValueType.distance,
                    leftRecordId_I,
                    rightRecordId_I, surveyorCount, cacheData
                )
                for (cycleJ in cycleI + 1 until cycleCount) {
                    val leftRecordId_J = IdUtil.createSurveyorRecordId(
                        stationId,
                        cycleJ,
                        point.pointId,
                        LeftOrRight.left
                    )
                    val rightRecordId_J = IdUtil.createSurveyorRecordId(
                        stationId,
                        cycleJ,
                        point.pointId,
                        LeftOrRight.right
                    )
                    val value2C_J = ValueOf2cMethod.valueOf2C(
                        SdAngleValueType.verAngle, leftRecordId_J, rightRecordId_J,
                        toleranceRecord.surveyorCount, cacheData
                    )
                    val distanceLeftAndRightAverageValue_J = AverageValueMethod.leftAndRightValue(
                        SdAngleValueType.distance,
                        leftRecordId_J,
                        rightRecordId_J, surveyorCount, cacheData
                    )
                    val verI = abs(value2C_I.value - value2C_J.value)
                    if (verCyclesIDifference < verI) {
                        verIError.add(
                            _ErrorItem(
                                point.pointName,
                                cycleI,
                                cycleJ,
                                verI
                            )
                        )
                    }

                    if (!point.aloneDistance) {
                        val distance =
                            abs(distanceLeftAndRightAverageValue_I.value - distanceLeftAndRightAverageValue_J.value)
                        if (distCyclesDifference < distance * 1000) {
                            distError.add(
                                _ErrorItem(
                                    point.pointName,
                                    cycleI,
                                    cycleJ,
                                    distance * 1000
                                )
                            )
                        }
                    }
                }
            }
        }

        return _errorString(
            horCyclesDirectValueDifference,
            verCyclesAngleDifference,
            verCyclesIDifference,
            distCyclesDifference,
            horValueError,
            verValueError,
            verIError,
            distError
        )

    }

    /**
     * 根据超限信息生成错误描述
     *
     * @param horValueError 水平角同方向值互差
     * @param verValueError 垂直角同方向值互差
     */
    private fun _errorString(
        horLimit: Double,
        verLimit: Double,
        verILimit: Double,
        distLimit: Double,
        horValueError: List<_ErrorItem>,
        verValueError: List<_ErrorItem>,
        verIError: List<_ErrorItem>,
        distError: List<_ErrorItem>
    ): String? {
        if (horValueError.isEmpty() && verValueError.isEmpty()) {
            return null
        }
        val stringBuilder = StringBuilder()
        if (horValueError.isNotEmpty()) {
            stringBuilder.append("测回间同方向值互差超限(限差:")
                .append(Util.formatDouble2String(horLimit, 1))
                .append("'')")
            for (item in horValueError) {
                stringBuilder
                    .append("\n   ")
                    .append(item.pointName)
                    .append(":\n")
                    .append("   |")
                    .append(SurveyorCycleUtil.cycleName(item.cycleIndex1))
                    .append("-")
                    .append(SurveyorCycleUtil.cycleName(item.cycleIndex2))
                    .append("| = ")
                    .append(Util.formatDouble2String(item.difference, 1))
                    .append("'')")
            }
        }

        if (verValueError.isNotEmpty()) {
            stringBuilder.append("\n测回间垂直角互差超限(限差:")
                .append(Util.formatDouble2String(verLimit, 1))
                .append("'')")
            for (item in verValueError) {
                stringBuilder
                    .append("\n   ")
                    .append(item.pointName)
                    .append(":\n")
                    .append("   |")
                    .append(SurveyorCycleUtil.cycleName(item.cycleIndex1))
                    .append("-")
                    .append(SurveyorCycleUtil.cycleName(item.cycleIndex2))
                    .append("| = ")
                    .append(Util.formatDouble2String(item.difference, 1))
                    .append("'')")
            }
        }

        if (verIError.isNotEmpty()) {
            stringBuilder.append("\n垂直角i指标同方向各测回较差超限(限差:")
                .append(Util.formatDouble2String(verILimit, 1))
                .append("'')")
            for (item in verIError) {
                stringBuilder
                    .append("\n   ")
                    .append(item.pointName)
                    .append(":\n")
                    .append("   |")
                    .append(SurveyorCycleUtil.cycleName(item.cycleIndex1))
                    .append("-")
                    .append(SurveyorCycleUtil.cycleName(item.cycleIndex2))
                    .append("| = ")
                    .append(Util.formatDouble2String(item.difference, 1))
                    .append("'')")
            }
        }

        if (distError.isNotEmpty()) {
            stringBuilder.append("\n测回间距离较差超限(限差:")
                .append(Util.formatDouble2String(distLimit, 1))
                .append("'')")
            for (item in distError) {
                stringBuilder
                    .append("\n   ")
                    .append(item.pointName)
                    .append(":\n")
                    .append("   |")
                    .append(SurveyorCycleUtil.cycleName(item.cycleIndex1))
                    .append("-")
                    .append(SurveyorCycleUtil.cycleName(item.cycleIndex2))
                    .append("| = ")
                    .append(Util.formatDouble2String(item.difference, 1))
                    .append("'')")
            }
        }

        return stringBuilder.toString()
    }
}