package com.bingce.controlnetwork.ui.widget.recyclerview;

import android.annotation.SuppressLint;
import android.view.MotionEvent;
import android.view.View;
import android.view.ViewGroup;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.recyclerview.widget.RecyclerView;

import com.bingce.controlapphelper.model.ViewItemType;

import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

import com.bingce.controlnetwork.R;
import com.bingce.controlnetwork.ui.widget.ViewFactory;
import com.bingce.controlnetwork.ui.widget.ViewFactory;

public abstract class AbstractDraggableRecycleViewAdapter<T>
        extends RecyclerView.Adapter<RecyclerView.ViewHolder>
        implements IDragSwipeChecker {
    private final List<T> items = new ArrayList<>();
    private @Nullable
    IViewHolderDragDelegate viewHolderDragDelegate;

    public AbstractDraggableRecycleViewAdapter(List<T> itemList) {
        if (itemList != null) {
            this.items.addAll(itemList);
        }
    }

    protected void replace(int start, List<T> data) {
        if (getItemCount() > start) {
            items.subList(start, getItemCount()).clear();
        }
        items.addAll(data);
    }

    protected int indexOf(T item) {
        return items.indexOf(item);
    }

    protected void remove(T item) {
        items.remove(item);
    }

    protected void removeAll(List<T> item) {
        items.removeAll(item);
    }

    protected void insert(int index, T item) {
        if (items.size() <= index) {
            return;
        }
        items.add(index, item);
    }

    protected void insertAll(List<T> item) {
        items.addAll(item);
    }

    protected void refreshAll(List<T> item) {
        items.clear();
        items.addAll(item);
        notifyDataSetChanged();
    }

    /**
     * 一井定向使用
     *
     * @param index 要插入的位置
     * @param item
     */
    protected void insertForWell(int index, List<T> item) {
        if (index < 0) {
            return;
        }
        if (index == items.size()) {
            items.addAll(item);
        } else if (index < items.size()) {
            items.addAll(index, item);
        }
    }

    final protected T getItem(int position) {
        if (position >= items.size()) {
            return null;
        }
        return items.get(position);
    }

    final protected void setItem(int position, T item) {
        if (position >= items.size() || position < 0) {
            return;
        }
        items.set(position, item);
    }

    public List<T> getItems() {
        return Collections.unmodifiableList(items);
    }

    public void setViewHolderDragDelegate(@Nullable IViewHolderDragDelegate viewHolderDragDelegate) {
        this.viewHolderDragDelegate = viewHolderDragDelegate;
    }

    @Override
    public int getItemViewType(int position) {
        return ViewItemType.draggableItem.ordinal();
    }

    @NonNull
    @Override
    public RecyclerView.ViewHolder onCreateViewHolder(@NonNull @NotNull ViewGroup parent, int viewType) {
        return Objects.requireNonNull(ViewFactory.generatorViewHolder(parent.getContext(), parent, ViewItemType.values()[viewType]));
    }

    @SuppressLint("ClickableViewAccessibility")
    @Override
    public void onBindViewHolder(@NonNull @NotNull RecyclerView.ViewHolder holder, int position) {
        bindItem(items.get(position), holder);
        View dragBar = holder.itemView.findViewById(R.id.drag_bar);
        if (dragBar != null) {
            dragBar.setOnTouchListener((v, event) -> {
                if (event.getAction() == MotionEvent.ACTION_DOWN) {
                    if (viewHolderDragDelegate != null) {
                        viewHolderDragDelegate.startDrag(holder);
                    }
                }
                return true;
            });
        }
        //原先代码注释
//        if (holder instanceof DraggableTextViewHolder) {
//            DraggableTextViewHolder draggableTextViewHolder = (DraggableTextViewHolder) holder;
//            draggableTextViewHolder.dragBar.setOnTouchListener((v, event) -> {
//                if (event.getAction() == MotionEvent.ACTION_DOWN) {
//                    if (viewHolderDragDelegate != null) {
//                        viewHolderDragDelegate.startDrag(holder);
//                    }
//                }
//                return true;
//            });
//        }
    }

    @Override
    public int getItemCount() {
        return items.size();
    }

    //交换
    public void onItemSwap(int from, int target) {
        T s = items.get(from);
        items.remove(from);
        items.add(target, s);
        notifyItemMoved(from, target);
    }

    //滑动
    public void onItemRemoved(int position) {
        items.remove(position);
        notifyItemRemoved(position);
    }

    protected abstract void bindItem(T item, RecyclerView.ViewHolder viewHolder);
}
